from typing import List

import torch
from e3nn import o3

from hpc2ml.data.nequlip_support.AtomicDataDict import ST
from ._graph_mixin import GraphModuleMixin


class Concat(GraphModuleMixin, torch.nn.Module):
    """Concatenate multiple fields into one."""

    def __init__(self, in_fields: List[str], out_field: str, irreps_in=None):
        super().__init__()
        if irreps_in is None:
            irreps_in = {}
        self.in_fields = list(in_fields)
        self.out_field = out_field
        self._init_irreps(irreps_in=irreps_in, required_irreps_in=self.in_fields)
        self.irreps_out[self.out_field] = sum(
            (self.irreps_in[k] for k in self.in_fields), o3.Irreps()
        )

    def forward(self, data: ST) -> ST:
        data[self.out_field] = torch.cat([data[k] for k in self.in_fields], dim=-1)
        return data



import warnings
from collections import deque

import numpy as np

from hpc2ml.optimize.basegpmin import BaseGPMin


class SimpleEIGPMin(BaseGPMin):
    """EffectiveImprovement-GPMin"""
    def __init__(self, atoms, **kwargs):
        super(SimpleEIGPMin, self).__init__(atoms, **kwargs)

    def step(self, f=None):

        atoms = self.atoms
        if f is None:
            f = atoms.get_forces()

        fc = self.force_consistent
        r0 = atoms.get_positions().reshape(-1)
        e0 = atoms.get_potential_energy(force_consistent=fc)
        self.update(r0, e0, f)

        r_add = None
        e_add = None
        f_add = None

        count = 0

        self.es = deque([e0], maxlen=3)

        while len(self.es) < 3 or np.std(self.es) > 0.01:  # 0.01 eV energy error

            r_, e_, f_ = self.resample_r_e_f(20,
                                             random_factor=0.01,  # this is fixed, 0.01 is best when test.
                                             )  # train

            if r_add is None:
                pass
            else:
                # Active learning iteration (add the sample with good performance).
                r_ = np.concatenate((r_, r_add), axis=0)
                e_ = np.concatenate((e_, e_add), axis=0)
                f_ = np.concatenate((f_, f_add), axis=0)

            self.update_batch(r_, e_, f_)  # train

            random_factor = max(0.05 * 1/(self.nsteps+1), 0.001)  # decrease the degree of position change
                                                                  # with step.

            r_ = self.resample_r(20, random_factor=random_factor)

            res = self.predict_mean_std_all(r_)

            ei = self.ei_balance(res)

            # ei = self.ei_balance(res, loss_coef=1-0.02*self.nsteps)
            # decrease the forces weight with step.
            # this should be just used for forces with
            index_ = np.argsort(-ei)[:5]
            r_add = r_[index_]

            r_add, e_add, f_add = self.r_to_e_f(r_add)

            index_min = np.argmin(e_add)  # best
            e1 = e_add[index_min]
            r1 = r_add[index_min]
            f1 = f_add[index_min]

            # if e1 < self.es[-1]+0.00001:
            if e1 < self.es[-1]:
                self.atoms.set_positions(r1.reshape(-1, 3))
                self.es.append(float(e1))

            self.function_calls += 1
            self.force_calls += 1

            if self.converged(f1.reshape(-1, 3)):
                break

            count += 1
            if count == 20:
                warnings.warn("A descent model could not be built")
                break

            # print(count, self.es[-1], e1)


        self.dump()






import torch


@torch.jit.script
def _poly_cutoff(x: torch.Tensor, factor: float, p: float = 6.0) -> torch.Tensor:
    x = x * factor

    out = 1.0
    out = out - (((p + 1.0) * (p + 2.0) / 2.0) * torch.pow(x, p))
    out = out + (p * (p + 2.0) * torch.pow(x, p + 1.0))
    out = out - ((p * (p + 1.0) / 2) * torch.pow(x, p + 2.0))

    return out * (x < 1.0)


class PolynomialCutoff(torch.nn.Module):
    _factor: float
    p: float

    def __init__(self, r_max: float, p: float = 6):
        r"""Polynomial cutoff, as proposed in DimeNet: https://arxiv.org/abs/2003.03123


        Parameters
        ----------
        r_max : float
            Cutoff radius

        p : int
            Power used in envelope function
        """
        super().__init__()
        assert p >= 2.0
        self.p = float(p)
        self._factor = 1.0 / float(r_max)

    def forward(self, x):
        """
        Evaluate cutoff function.

        x: torch.Tensor, input distance
        """
        return _poly_cutoff(x, self._factor, p=self.p)



# -*- coding: utf-8 -*-
# @Time  : 2023/4/19 17:10
# @Author : boliqq07
# @Software: PyCharm
# @License: MIT License

import copy
from typing import Tuple, List

import ase
from ase import Atoms
from ase.spacegroup import get_spacegroup

from perovskite.core.general_rotate import rotate_times, rotate_by_angle
from perovskite.core.general_standard import perovskite_prim_221, perovskite_conv_221, \
    perovskite_prim_62, perovskite_conv_127, perovskite_prim_127, perovskite_conv_62
from perovskite.core.general_substitute import sub_atoms
from perovskite.core.rotate_substitude import SubstituteWithRotatedB


class Perovskite(Atoms):
    """Specialized custom classes for Perovskite materials."""

    def __init__(self, symbols=None,
                 positions=None, numbers=None,
                 tags=None, momenta=None, masses=None,
                 magmoms=None, charges=None,
                 scaled_positions=None,
                 cell=None, pbc=None, celldisp=None,
                 constraint=None,
                 calculator=None,
                 info=None,
                 velocities=None):
        super().__init__(symbols=symbols,
                         positions=positions, numbers=numbers,
                         tags=tags, momenta=momenta, masses=masses,
                         magmoms=magmoms, charges=charges,
                         scaled_positions=scaled_positions,
                         cell=cell, pbc=pbc, celldisp=celldisp,
                         constraint=constraint,
                         calculator=calculator,
                         info=info,
                         velocities=velocities)

    @classmethod
    def from_atoms(cls, atoms: Atoms):
        """From Atoms to Perovskite."""
        atomss = cls(cell=atoms.cell, pbc=atoms.pbc, info=atoms.info,
                     celldisp=atoms._celldisp.copy())

        atomss.arrays = {}
        for name, a in atoms.arrays.items():
            atomss.arrays[name] = a.copy()
        atomss.constraints = copy.deepcopy(atoms.constraints)
        return atomss

    def to_atoms(self) -> Atoms:
        """From Perovskite to Atoms."""
        atoms = Atoms(cell=self.cell, pbc=self.pbc, info=self.info,
                      celldisp=self._celldisp.copy())

        atoms.arrays = {}
        for name, a in self.arrays.items():
            atoms.arrays[name] = a.copy()
        atoms.constraints = copy.deepcopy(self.constraints)
        return atoms

    def rotate_by_angle(self, angle: int = 90, axis: Tuple[Tuple, int] = ((1, 0, 0),), center=(0, 0, 0)) -> List[
        Atoms]:
        """Rotate atoms self by angle along x, y, z axis"""
        return rotate_by_angle(self.copy(), angle, axis, center=center)

    def rotate_by_times(self, times: int = 4, axis: Tuple[Tuple, int, str] = ((1, 0, 0),)) -> List[Atoms]:
        """Rotate atoms self by times along x, y, z axis"""
        return rotate_times(self.copy(), times, axis)

    @classmethod
    def from_standard(cls, template, a_atom="Ca", b_atom="Ti", c_atom="O", a=7, alpha=1.0,
                      move=0.04, cycle=0.04, size=(1, 1, 1), **kwargs) -> "Perovskite":
        """
        Get perovskite by standard.

        Args:
            template: (str), perovskite type, available in ("prim_221","conv_221","prim_127","conv_127","prim_62","conv_62")
            a_atom: (str), a atom
            b_atom: (str), b atom
            c_atom: (str), c atom
            a: (str), a axis length of cell
            alpha: (float), c / a or b / a axis ratio
            move: (float), distortion of tetrahedron
            cycle: (float), distortion of tetrahedron
            size: (tuple), supercell site, default 111
            **kwargs:

        Returns:
            perovskite: (Perovskite), new object.

        """

        mp = {
            "prim_221": perovskite_prim_221,
            "conv_221": perovskite_conv_221,
            "prim_127": perovskite_prim_127,
            "conv_127": perovskite_conv_127,
            "prim_62": perovskite_prim_62,
            "conv_62": perovskite_conv_62
        }

        func = mp[template]
        atoms = func(a_atom=a_atom, b_atom=b_atom, c_atom=c_atom, a=a, alpha=alpha,
                     move=move, cycle=cycle, size=size, **kwargs)
        return cls(symbols=None, positions=atoms.positions, numbers=atoms.numbers,
                   cell=atoms.cell, pbc=atoms.pbc, celldisp=atoms._celldisp,
                   calculator=atoms.calc, )

    def sub_atoms(self, a: str, b: [str, ase.Atom, ase.Atoms], index=None) -> List[Atoms]:
        """Rotate atoms by times along x, y, z axis"""
        return sub_atoms(self.copy(), a, b, index)

    def substitute_with_rotated_b(self, old: str, new: [str, ase.Atom, ase.Atoms], index=None,
                                  strategy="auto", space_group_number=None, symprec=1e-5,
                                  ) -> "Perovskite":
        """
        Substitute atoms in atoms.

        such as: ABX3 -> ACX3

        such as: ABX3 -> ABO3

        such as: ABX3 -> (CH3NH3)BX3


        Args:
            old: (str,) old atom name.
            new: (str, ase.Atom, ase.Atoms), new atoms.
            index: tuple, part of index of all `a` atom (from 0 to N_a - 1). if index is None, substitute all `a` atom.
            strategy: ("same", "mirror", "random", "auto"), strategy to deal with b atoms before substitute.
            space_group_number: (float, "auto"), space_group_number
            symprec: float, tolerance to find space_group_number, if space_group_number="auto"

        Returns:


        """
        if space_group_number is None:
            sg = get_spacegroup(self, symprec=symprec)
            space_group_number = sg.no

        swrb = SubstituteWithRotatedB(strategy=strategy, space_group_number=space_group_number)

        return swrb.sub_atoms(self.copy(), a=old, b=new, index=index)

    def to_structure(self):
        from pymatgen.io.ase import AseAtomsAdaptor
        aaa = AseAtomsAdaptor()
        return aaa.get_structure(self)

    def show(self):
        """
        Plot by ase.
        """
        from ase.visualize import view
        view(self)

    def view(self):
        """The same as show. plot by ase."""
        self.show()

    @staticmethod
    def T_f(rA, rB, rX):
        """
        公差系数
        """
        t_f = (rA + rX) / (1.414 * (rB + rX))
        return t_f

    @staticmethod
    def O_f(rB, rX):
        """
        八面体因子
        """
        o_f = rB / rX
        return o_f


# import logging
# from typing import Optional
#
# from e3nn import o3
#
# from hpc2ml.data import AtomicDataDict
# from ..nn import (
#     SequentialGraphNetwork,
#     AtomwiseLinear,
#     AtomwiseReduce,
#     ConvNetLayer,
# )
# from ..nn.embedding import (
#     OneHotAtomEncoding,
#     RadialBasisEdgeEncoding,
#     SphericalHarmonicEdgeAttrs,
# )
#
#
# # from . import builder_utils
#
#
# def SimpleIrrepsConfig(config, prefix: Optional[str] = None):
#     """Builder that pre-processes options to allow "simple" configuration of irreps."""
#
#     # We allow some simpler parameters to be provided, but if they are,
#     # they have to be correct and not overridden
#     simple_irreps_keys = ["l_max", "parity", "num_features"]
#     real_irreps_keys = [
#         "chemical_embedding_irreps_out",
#         "feature_irreps_hidden",
#         "irreps_edge_sh",
#         "conv_to_output_hidden_irreps_out",
#     ]
#
#     prefix = "" if prefix is None else f"{prefix}_"
#
#     has_simple: bool = any(
#         (f"{prefix}{k}" in config) or (k in config) for k in simple_irreps_keys
#     )
#     has_full: bool = any(
#         (f"{prefix}{k}" in config) or (k in config) for k in real_irreps_keys
#     )
#     assert has_simple or has_full
#
#     update = {}
#     if has_simple:
#         # nothing to do if not
#         lmax = config.get(f"{prefix}l_max", config["l_max"])
#         parity = config.get(f"{prefix}parity", config["parity"])
#         num_features = config.get(f"{prefix}num_features", config["num_features"])
#         update[f"{prefix}chemical_embedding_irreps_out"] = repr(
#             o3.Irreps([(num_features, (0, 1))])  # n scalars
#         )
#         update[f"{prefix}irreps_edge_sh"] = repr(
#             o3.Irreps.spherical_harmonics(lmax=lmax, p=-1 if parity else 1)
#         )
#         update[f"{prefix}feature_irreps_hidden"] = repr(
#             o3.Irreps(
#                 [
#                     (num_features, (l, p))
#                     for p in ((1, -1) if parity else (1,))
#                     for l in range(lmax + 1)
#                 ]
#             )
#         )
#         update[f"{prefix}conv_to_output_hidden_irreps_out"] = repr(
#             # num_features // 2  scalars
#             o3.Irreps([(max(1, num_features // 2), (0, 1))])
#         )
#
#     # check update is consistant with config
#     # (this is necessary since it is not possible
#     #  to delete keys from config, so instead of
#     #  making simple and full styles mutually
#     #  exclusive, we just insist that if full
#     #  and simple are provided, full must be
#     #  consistant with simple)
#     for k, v in update.items():
#         if k in config:
#             assert (
#                     config[k] == v
#             ), f"For key {k}, the full irreps options had value `{config[k]}` inconsistant with the value derived from the simple irreps options `{v}`"
#         config[k] = v
#
#
# def EnergyModel(
#         config,
# ) -> SequentialGraphNetwork:
#     """Base default energy model archetecture.
#
#     For minimal and full configuration option listings, see ``minimal.yaml`` and ``example.yaml``.
#     """
#     logging.debug("Start building the network model")
#
#     # builder_utils.add_avg_num_neighbors(
#     #     config=config, initialize=initialize, dataset=dataset
#     # )
#
#     num_layers = config.get("num_layers", 3)
#
#     layers = {
#         # -- Encode --
#         "one_hot": OneHotAtomEncoding,
#         "spharm_edges": SphericalHarmonicEdgeAttrs,
#         "radial_basis": RadialBasisEdgeEncoding,
#         # -- Embed features --
#         "chemical_embedding": AtomwiseLinear,
#     }
#
#     # add convnet layers
#     # insertion preserves order
#     for layer_i in range(num_layers):
#         layers[f"layer{layer_i}_convnet"] = ConvNetLayer
#
#     # .update also maintains insertion order
#     layers.update(
#         {
#             # TODO: the next linear throws out all L > 0, don't create them in the last layer of convnet
#             # -- output block --
#             "conv_to_output_hidden": AtomwiseLinear,
#             "output_hidden_to_scalar": (
#                 AtomwiseLinear,
#                 dict(irreps_out="1x0e", out_field=AtomicDataDict.PER_ATOM_ENERGY_KEY),
#             ),
#         }
#     )
#
#     layers["total_energy_sum"] = (
#         AtomwiseReduce,
#         dict(
#             reduce="sum",
#             in_field=AtomicDataDict.PER_ATOM_ENERGY_KEY,
#             out_field=AtomicDataDict.TOTAL_ENERGY_KEY,
#         ),
#     )
#
#     return SequentialGraphNetwork.from_parameters(
#         shared_params=config,
#         layers=layers,
#     )



# -*- coding: utf-8 -*-
# @Time  : 2022/9/27 15:12
# @Author : boliqq07
# @Software: PyCharm
# @License: MIT License

import torch
from matplotlib import pyplot as plt
from torch import nn
from torch.nn import ReLU


class Morse(torch.nn.Module):
    """Morse potential"""

    def __init__(self):
        super().__init__()
        from ase.data import covalent_radii
        radii = torch.from_numpy(covalent_radii).float()
        self.lin1 = nn.Sequential(
            nn.Linear(2, 8, bias=True), ReLU(),
            nn.Linear(8, 8, bias=True), ReLU(),
            nn.Linear(8, 8, bias=True)
        )
        self.register_buffer('radii', radii)

    def forward(self, cen, nei, r):
        cen = cen.view(-1)
        nei = nei.view(-1)
        if r.ndim == 1 or (r.ndim == 2 and r.shape[1] == 1):
            r = r.view(-1, 1)
            return self._forward_r(cen, nei, r)
        elif r.ndim == 2 and r.shape[1] == 3:
            return self._forward_xyz(cen, nei, r)
        else:
            raise NotImplementedError("r are just accept distance (N,) or (N,1) or distance_vec (N,3)")

    def _forward_r(self, cen, nei, r):

        r0 = (self.radii[nei] + self.radii[cen]).view(-1, 1)
        param_in = torch.vstack((nei + cen, nei * cen)).T.float()
        param_out = self.lin1(param_in)
        # param_out = torch.abs(param_out) # not use

        D, k, c, alpha, n_3, n1, n2, _ = torch.chunk(param_out, chunks=8, dim=1)

        D = torch.abs(D)
        k = torch.abs(k)
        n1 = torch.abs(n1)

        n1 = 0.001 * n1
        n2 = 0.001 * n2
        n_3 = 0.001 * n_3

        alpha = torch.minimum(alpha, torch.full_like(alpha, 20))
        alpha = torch.maximum(alpha, torch.full_like(alpha, -20))
        k = torch.minimum(k, torch.full_like(alpha, 20))

        MV_COEF = 1.0
        LINEBASE_COEF = 2.0

        alpha = (1 + 0.01 * alpha) * MV_COEF
        k = (0.8 + 0.01 * k) * LINEBASE_COEF

        core = -k * (r - alpha * r0)

        pot = D * torch.pow(1 - torch.exp(core), 2) + n_3 * r ** (1 / 3) + n2 * r ** 2

        pot = torch.minimum(pot, torch.full_like(pot, 10))

        pot = pot - n1 * r + c

        return pot

    def _forward_xyz(self, cen, nei, r):

        r1 = torch.sum(r ** 2, dim=1, keepdim=True)

        pot = self._forward_r(cen, nei, r1)

        r2 = r * torch.abs(r)

        beta = r2 / r1

        pot = beta * pot

        return pot


if __name__ == "__main__":
    mos = Morse()

    # 0.31,  # H
    # 0.66,  # O

    r1 = torch.arange(0.2, 5, step=0.02, requires_grad=True)

    a = torch.full_like(r1, 1, dtype=torch.long)
    b = torch.full_like(r1, 8, dtype=torch.long)

    r = r1
    lj = Morse()
    res = lj(a, b, r)
    re = res.mean()
    re.backward()

    nps = res.detach().numpy()
    r = r.detach().numpy()
    plt.plot(r, nps)
    plt.show()

    r1 = torch.arange(0.10, 1.7, step=0.02, requires_grad=True)
    r2 = torch.arange(0.10, 1.7, step=0.02, requires_grad=True)
    r3 = torch.arange(0.10, 1.7, step=0.02, requires_grad=True)

    r = torch.vstack((r1, r2, r3)).T
    a = torch.full_like(r1, 1, dtype=torch.long)
    b = torch.full_like(r1, 8, dtype=torch.long)

    r = r
    lj = Morse()
    res = lj(a, b, r)
    re = res.mean()
    re.backward()

    nps = res.detach().numpy()

    r = r.detach().numpy()
    plt.plot(r, nps[:, 0])
    plt.show()

    plt.plot(r, nps[:, 1])
    plt.show()

    plt.plot(r, nps[:, 2])
    plt.show()



from ._atomwise import (  # noqa: F401
    AtomwiseOperation,
    AtomwiseReduce,
    AtomwiseLinear,
    PerSpeciesScaleShift,
)  # noqa: F401
from ._concat import Concat  # noqa: F401
from ._convnetlayer import ConvNetLayer  # noqa: F401
from ._grad_output import GradientOutput, PartialForceOutput, StressOutput  # noqa: F401
from ._graph_mixin import GraphModuleMixin, SequentialGraphNetwork  # noqa: F401
from ._interaction_block import InteractionBlock  # noqa: F401
from ._rescale import RescaleOutput  # noqa: F401
from ._util import SaveForOutput  # noqa: F401



# -*- coding: utf-8 -*-

# @Time    : 2021/7/15 22:34
# @Email   : 986798607@qq.com
# @Software: PyCharm
# @License: BSD 3-Clause


import copy
from typing import List

import ase
import numpy as np
from ase import Atoms


class CircaAtoms(Atoms):
    """check the 2 atoms is similar."""

    @classmethod
    def from_atoms(cls, atoms: ase.Atoms):
        """Return a copy."""
        atomss = cls(cell=atoms.cell, pbc=atoms.pbc, info=atoms.info,
                     celldisp=atoms._celldisp.copy())

        atomss.arrays = {}
        for name, a in atoms.arrays.items():
            atomss.arrays[name] = a.copy()
        atomss.constraints = copy.deepcopy(atoms.constraints)
        return atomss

    def to_atoms(self) -> Atoms:
        """Return a copy."""
        atoms = Atoms(cell=self.cell, pbc=self.pbc, info=self.info,
                      celldisp=self._celldisp.copy())

        atoms.arrays = {}
        for name, a in self.arrays.items():
            atoms.arrays[name] = a.copy()
        atoms.constraints = copy.deepcopy(self.constraints)
        return atoms

    def __eq__(self, other, tol=2):

        if not isinstance(other, Atoms):
            return False

        a = self.arrays
        b = other.arrays

        poa = np.round(a['positions'], decimals=tol)
        pob = np.round(b['positions'], decimals=tol)

        poa = np.sort(poa, axis=0)
        poa_index = np.argsort(poa, axis=0)
        pob = np.sort(pob, axis=0)
        pob_index = np.argsort(pob, axis=0)
        return (len(self) == len(other) and
                (poa == pob).all()
                and (a['numbers'][poa_index] == b['numbers'][pob_index]).all())


def filter_dup(atomss: List[Atoms, CircaAtoms]) -> List[Atoms]:
    """Remove the same atoms."""
    temp = []
    for n, i in enumerate(atomss):
        mark_i = []
        for j in temp:
            if i == j:
                mark_i.append(True)
            else:
                mark_i.append(False)
        if not any(mark_i):
            temp.append(i)
    return temp



"""For quick use.

Examples:
    >>> atoms = aaa.get_atoms(structure=structure)
    >>> structure = aaa.get_structure(atoms=atoms)
"""

from pymatgen.io.ase import AseAtomsAdaptor

aaa = AseAtomsAdaptor()



from typing import Sequence, List, Union

import torch
from e3nn.util.jit import compile_mode

from hpc2ml.data.nequlip_support.AtomicDataDict import ST
from ._graph_mixin import GraphModuleMixin


@compile_mode("script")
class RescaleOutput(GraphModuleMixin, torch.nn.Module):
    """Wrap a model and rescale its outputs when in ``eval()`` mode.

    Args:
        model : GraphModuleMixin
            The model whose outputs are to be rescaled.
        scale_keys : list of keys, default []
            Which fields to rescale.
        shift_keys : list of keys, default []
            Which fields to shift after rescaling.
        related_scale_keys: list of keys that could be contingent to this rescale
        related_shift_keys: list of keys that could be contingent to this rescale
        scale_by : floating or Tensor, default 1.
            The scaling factor by which to multiply fields in ``scale``.
        shift_by : floating or Tensor, default 0.
            The shift to add to fields in ``shift``.
        irreps_in : dict, optional
            Extra inputs expected by this beyond those of `model`; this is only present for compatibility.
    """

    scale_keys: List[str]
    shift_keys: List[str]
    related_scale_keys: List[str]
    related_shift_keys: List[str]
    scale_trainble: bool
    rescale_trainable: bool

    has_scale: bool
    has_shift: bool

    def __init__(
            self,
            model: GraphModuleMixin,
            scale_keys: Union[Sequence[str], str] = [],
            shift_keys: Union[Sequence[str], str] = [],
            related_shift_keys: Union[Sequence[str], str] = [],
            related_scale_keys: Union[Sequence[str], str] = [],
            scale_by=None,
            shift_by=None,
            shift_trainable: bool = False,
            scale_trainable: bool = False,
            irreps_in: dict = {},
    ):
        super().__init__()

        self.model = model
        scale_keys = [scale_keys] if isinstance(scale_keys, str) else scale_keys
        shift_keys = [shift_keys] if isinstance(shift_keys, str) else shift_keys
        all_keys = set(scale_keys).union(shift_keys)

        # Check irreps:
        for k in irreps_in:
            if k in model.irreps_in and model.irreps_in[k] != irreps_in[k]:
                raise ValueError(
                    f"For in_field '{k}', the provided explicit `irreps_in` ('{k}': {irreps_in[k]}) are incompataible with those of the wrapped `model` ('{k}': {model.irreps_in[k]})"
                )
        for k in all_keys:
            if k not in model.irreps_out:
                raise KeyError(
                    f"Asked to scale or shift '{k}', but '{k}' is not in the outputs of the provided `model`."
                )
        for k in shift_keys:
            if model.irreps_out[k] is not None and model.irreps_out[k].lmax > 0:
                raise ValueError(
                    f"It doesn't make sense to shift non-scalar target '{k}'."
                )

        irreps_in.update(model.irreps_in)
        self._init_irreps(irreps_in=irreps_in, irreps_out=model.irreps_out)

        self.scale_keys = list(scale_keys)
        self.shift_keys = list(shift_keys)
        self.related_scale_keys = list(set(related_scale_keys).union(scale_keys))
        self.related_shift_keys = list(set(related_shift_keys).union(shift_keys))

        self.has_scale = scale_by is not None
        self.scale_trainble = scale_trainable
        if self.has_scale:
            scale_by = torch.as_tensor(scale_by)
            if self.scale_trainble:
                self.scale_by = torch.nn.Parameter(scale_by)
            else:
                self.register_buffer("scale_by", scale_by)
        elif self.scale_trainble:
            raise ValueError(
                "Asked for a scale_trainable, but this RescaleOutput has no scaling (`scale_by = None`)"
            )
        else:
            # register dummy for TorchScript
            self.register_buffer("scale_by", torch.Tensor())

        self.has_shift = shift_by is not None
        self.rescale_trainable = shift_trainable
        if self.has_shift:
            shift_by = torch.as_tensor(shift_by)
            if self.rescale_trainable:
                self.shift_by = torch.nn.Parameter(shift_by)
            else:
                self.register_buffer("shift_by", shift_by)
        elif self.rescale_trainable:
            raise ValueError(
                "Asked for a shift_trainable, but this RescaleOutput has no shift (`shift_by = None`)"
            )
        else:
            # register dummy for TorchScript
            self.register_buffer("shift_by", torch.Tensor())

        # Finally, we tell all the modules in the model that there is rescaling
        # This allows them to update parameters, like physical constants with units,
        # that need to be scaled

        # Note that .modules() walks the full tree, including self
        for mod in self.get_inner_model().modules():
            if isinstance(mod, GraphModuleMixin):
                callback = getattr(mod, "update_for_rescale", None)
                if callable(callback):
                    # It gets the `RescaleOutput` as an argument,
                    # since that contains all relevant information
                    callback(self)

    def get_inner_model(self):
        """Get the outermost child module that is not another ``RescaleOutput``"""
        model = self.model
        while isinstance(model, RescaleOutput):
            model = model.model
        return model

    def forward(self, data: ST) -> ST:
        data = self.model(data)
        if self.training:
            return data
        else:
            # Scale then shift
            if self.has_scale:
                for in_field in self.scale_keys:
                    data[in_field] = data[in_field] * self.scale_by
            if self.has_shift:
                for in_field in self.shift_keys:
                    data[in_field] = data[in_field] + self.shift_by
            return data

    @torch.jit.export
    def scale(
            self,
            data: ST,
            force_process: bool = False,
    ) -> ST:
        """Apply rescaling to ``data``, in place.

        Only processes the data if the module is in ``eval()`` mode, unless ``force_process`` is ``True``.

        Args:
            data (map-like): a dict, ``AtomicDataDict``, ``AtomicData``, ``torch_geometric.data.Batch``, or anything else dictionary-like
            force_process (bool): if ``True``, scaling will be done regardless of whether the model is in train or evaluation mode.
        Returns:
            ``data``, modified in place
        """
        data = data.copy()
        if self.training and not force_process:
            return data
        else:
            if self.has_scale:
                for in_field in self.scale_keys:
                    if in_field in data:
                        data[in_field] = data[in_field] * self.scale_by
            if self.has_shift:
                for in_field in self.shift_keys:
                    if in_field in data:
                        data[in_field] = data[in_field] + self.shift_by
            return data

    @torch.jit.export
    def unscale(
            self,
            data: ST,
            force_process: bool = False,
    ) -> ST:
        """Apply the inverse of the rescaling operation to ``data``, in place.

        Only processes the data if the module is in ``train()`` mode, unless ``force_process`` is ``True``.

        Args:
            data (map-like): a dict, ``AtomicDataDict``, ``AtomicData``, ``torch_geometric.data.Batch``, or anything else dictionary-like
            force_process (bool): if ``True``, unscaling will be done regardless of whether the model is in train or evaluation mode.
        Returns:
            ``data``
        """
        data = data.copy()
        if self.training or force_process:
            # To invert, -shift then divide by scale
            if self.has_shift:
                for in_field in self.shift_keys:
                    if in_field in data:
                        data[in_field] = data[in_field] - self.shift_by
            if self.has_scale:
                for in_field in self.scale_keys:
                    if in_field in data:
                        data[in_field] = data[in_field] / self.scale_by
            return data
        else:
            return data



# -*- coding: utf-8 -*-

from typing import Tuple, Dict, List, Union, Callable, Mapping

import numpy as np
import pandas as pd
import torch
from ase import Atoms
from ase.calculators.singlepoint import SinglePointCalculator
from ase.stress import full_3x3_to_voigt_6_stress
from pymatgen.core import Structure, Lattice, Molecule
from torch_geometric.data import Batch
from torch_geometric.data import Data
from torch_geometric.data.collate import collate
from torch_geometric.data.data import BaseData
from torch_geometric.data.in_memory_dataset import nested_iter
from torch_geometric.data.separate import separate

from hpc2ml.data.io import aaa
from hpc2ml.data.io.main import sparse_source_data
from hpc2ml.data.structuretodata import StructureToData


class MtBatchData(Batch):
    """
    One simple Batch data (torch_geometrics) integrate with generation and statistics,
    which is one higher wrapper of StructureToData.

    Examples:
        >>> mtdata = MtBatchData.from_data_list(batch)
        >>> data_list = MtBatchData.to_data_list()

    Examples:
        >>> mtdata = MtBatchData.from_batch(batch)
        >>> batch = MtBatchData.to_batch()

    Examples:
        >>> mtdata = MtBatchData.from_data(data,slices)
        >>> data,slices = MtBatchData.to_data()

    Examples:
        >>> mtdata = MtBatchData.from_atoms(atoms)
        >>> atoms = MtBatchData.to_atoms()

    Examples:
        >>> mtdata = MtBatchData.from_structure(structure)
        >>> structure = MtBatchData.to_structure()

    Examples:
        >>> mtdata = MtBatchData.from_sparse_source_data(...)
    See Also: StructureToData.

    Examples:
        >>> mtdata = MtBatchData.from_local_data_dict(file1, convert=StructureToData(),
        ... msg_name=(), save=False, store_root_dir=".",)

    """

    @classmethod
    def from_batch(cls, batch_data: Batch) -> "MtBatchData":
        """From torch_geometrics's Batch."""
        res = [batch_data.get_example(i) for i in range(batch_data.num_graphs)]
        return cls.from_data_list(res)

    def to_batch(self) -> Batch:
        """To torch_geometrics's Batch."""
        res = [self.get_example(i) for i in range(self.num_graphs)]
        return Batch.from_data_list(res)

    @staticmethod
    def _len(slices) -> int:
        if slices is None:
            return 1
        for _, value in nested_iter(slices):
            return len(value) - 1
        return 0

    @classmethod
    def from_data(cls, data: Data, slices) -> "MtBatchData":
        """From torch_geometrics's Data and slices (dict)."""
        l = cls._len(slices)
        res = [separate(
            cls=data.__class__,
            batch=data,
            idx=idx,
            slice_dict=slices,
            decrement=False, ) for idx in range(l)]
        return cls.from_data_list(res)

    def to_data(self) -> Tuple[BaseData, Mapping]:
        """To torch_geometrics's Data and slices (dict)."""
        res = [self.get_example(i) for i in range(self.num_graphs)]

        data, slices, _ = collate(
            self[0].__class__,
            data_list=res,
            increment=False,
            add_batch=False,
        )

        return data, slices

    def add_prop(self, key, value, slice_type="atom"):
        """Add properties could be sliced."""
        if slice_type in ["atom", "node"]:
            ci = self._inc_dict["x"]
            si = self._slice_dict["x"]
        elif slice_type == "sample":
            ci = self._inc_dict["natoms"]
            si = self._slice_dict["natoms"]
        elif slice_type == "edge":
            ci = self._inc_dict["edge_index"]
            si = self._slice_dict['edge_index']
        else:
            raise NotImplementedError

        setattr(self, key, value)
        self.keys.append(key)
        self._slice_dict.update({key:si})
        self._inc_dict.update({key:ci})

    @classmethod
    def from_atoms(cls, atoms: Union[Atoms, List[Atoms]], convert=StructureToData(),
                   msg_name=None, remain_old_edge=True, **kwargs, ) -> "MtBatchData":
        """From ase atoms."""
        if msg_name is None:
            msg_name = ("energy", "energies", "forces", "stress")

        if isinstance(atoms, Atoms):
            atoms = [atoms, ]
        else:
            atoms = atoms

        convert.just_predefined = False

        add_kw = {name: [i.calc.results[name] if i.calc is not None and name in i.calc.results else None for i in atoms]
                  for name in msg_name}

        if remain_old_edge:
            msg_name2 = ("edge_index",)
            add_kw2 = {name: [i.info[name] if name in i.info else None for i in atoms] for name in msg_name2}
            add_kw.update(add_kw2)

        add_kw.update(kwargs)
        kwargs = {k: v for k, v in add_kw.items() if v[0] is not None}

        st = [aaa.get_structure(i) for i in atoms]
        return cls.from_structure(structure=st, convert=convert, **kwargs, )

    def to_atoms(self, extra_fields=None, ) -> Union[List[Atoms], Atoms]:
        """Build a (list of) ``ase.Atoms`` object(s) from an ``AtomicData`` object.

        For each unique batch number provided in ``AtomicDataDict.BATCH_KEY``,
        an ``ase.Atoms`` object is created. If ``AtomicDataDict.BATCH_KEY`` does not
        exist in self, a single ``ase.Atoms`` object is created.

        Args:
            extra_fields: fields other than those handled explicitly (currently
                those defining the structure as well as energy, per-atom energy,
                and forces) to include in the output object. Per-atom (per-node)
                quantities will be included in ``arrays``; per-graph and per-edge
                quantities will be included in ``info``.

        Returns:
            A list of ``ase.Atoms`` objects if ``AtomicDataDict.BATCH_KEY`` is in self
            and is not None. Otherwise, a single ``ase.Atoms`` object is returned.
        """
        if extra_fields is None:
            # extra_fields = []
            extra_fields = ["edge_index", ]

        positions = getattr(self, "pos")

        if positions.device != torch.device("cpu"):
            raise TypeError(
                "Explicitly move this `AtomicData` to CPU using `.to()` before calling `to_ase()`.")

        batch = getattr(self, "batch")
        n_batches = batch.max() + 1
        batch_atoms = []
        for batch_idx in range(n_batches):

            data = self.get_example(batch_idx)
            pbc = getattr(data, "pbc", None)
            atomic_nums = getattr(data, "z", None)
            pos = getattr(data, "pos", None)
            cell = getattr(data, "cell", None)
            edge_index = getattr(data, "edge_index", None)

            energy = getattr(data, "energy", None)
            energies = getattr(data, "energies", None)
            forces = getattr(data, "forces", None)
            stress = getattr(data, "stress", None)

            mol = Atoms(
                numbers=atomic_nums.cpu().numpy().ravel(),  # must be flat for ASE
                positions=pos.cpu().numpy(),
                cell=torch.squeeze(cell).cpu().numpy() if cell is not None else None,
                pbc=pbc.view(-1).cpu().numpy() if pbc is not None else None, )

            # add other information
            fields = {}
            if energies is not None:
                fields["energies"] = energies.cpu().numpy()
            if energy is not None:
                fields["energy"] = energy.cpu().numpy()
            if forces is not None:
                fields["forces"] = forces.cpu().numpy()
            if stress is not None:
                stress = stress.cpu().numpy()
                if self["stress"].shape[1] == 3:
                    fields["stress"] = full_3x3_to_voigt_6_stress(stress)
                else:
                    fields["stress"] = stress

            mol.calc = SinglePointCalculator(mol, **fields)

            # add other information
            for key in extra_fields:
                if key == "edge_index":
                    mol.info["edge_index"] = edge_index.cpu().numpy().astype(np.long)
                else:
                    raise RuntimeError(
                        f"Extra field `{key}` isn't registered as node/edge/graph"
                    )

            batch_atoms.append(mol)

        if batch is not None:
            return batch_atoms
        else:
            assert len(batch_atoms) == 1
            return batch_atoms[0]

    @classmethod
    def from_sparse_source_data(cls, source_file="vasprun.xml", source_path="./pure_opt",
                                fmt="vasprun_traj", convert=StructureToData(),
                                msg_name=(), store=False, store_root_dir=".", **kwargs) -> "MtBatchData":
        """
        Sparse data by different function.

        Args:
            store: bool, save.
            store_root_dir:str, store path.
            source_file: file name to sparse.
            source_path: (str,list), path or paths.
            fmt: str, load function named  "sparse_{fmt}" in from ``hpc2ml.data.io`` .
            n_jobs: int, the parallel number to load,
            msg_name: tuple, names to convert.
            convert: StructureToData, converter.
            **kwargs: dict, the parameter in "sparse_{fmt}".

        Returns:
            data_dict:dict, data
        """
        dicts = sparse_source_data(source_file=source_file, source_path=source_path, fmt=fmt,
                                   **kwargs)
        convert.data_dict = dicts
        if store:
            res = convert.transform_data_dict_and_save(msg_name=msg_name, store_root_dir=store_root_dir, )
        else:
            res = convert.transform_data_dict(msg_name=msg_name)
        return cls.from_data_list(res)

    @classmethod
    def from_local_data_dict(cls, file: Union[Dict, str], convert=StructureToData(),
                             msg_name=(), save=False, store_root_dir=".", ) -> "MtBatchData":
        """From dict data dumped by pandas, in local disk"""
        if isinstance(file, str):
            dicts = pd.read_pickle(file)
        else:
            dicts = file

        convert.data_dict = dicts
        if save:
            res = convert.transform_data_dict_and_save(msg_name=msg_name, store_root_dir=store_root_dir, )
        else:
            res = convert.transform_data_dict(msg_name=msg_name)
        return cls.from_data_list(res)

    @classmethod
    def from_structure(cls, structure: Union[Structure, List[Structure]], convert=StructureToData(),
                       **kwargs, ) -> "MtBatchData":
        """From pymatgen Structure"""
        if isinstance(structure, Structure):
            structure = [structure, ]
        convert.just_predefined = False
        res = convert.transform(structure=structure, **kwargs)
        return cls.from_data_list(res)

    def to_structure(self, extra_fields=None, ) -> Union[Tuple[List[Structure], Dict], Tuple[Structure, Dict]]:
        """Build a (list of) ``ase.Atoms`` object(s) from an ``AtomicData`` object.

        For each unique batch number provided in ``AtomicDataDict.BATCH_KEY``,
        an ``ase.Atoms`` object is created. If ``AtomicDataDict.BATCH_KEY`` does not
        exist in self, a single ``ase.Atoms`` object is created.

        Args:
            extra_fields: fields other than those handled explicitly (currently
                those defining the structure as well as energy, per-atom energy,
                and forces) to include in the output object. Per-atom (per-node)
                quantities will be included in ``arrays``; per-graph and per-edge
                quantities will be included in ``info``.

        Returns:
            A list of ``ase.Atoms`` objects if ``AtomicDataDict.BATCH_KEY`` is in self
            and is not None. Otherwise, a single ``ase.Atoms`` object is returned.
        """
        if extra_fields is None:
            extra_fields = ["edge_index", ]

        positions = getattr(self, "pos")

        if positions.device != torch.device("cpu"):
            raise TypeError(
                "Explicitly move this `AtomicData` to CPU using `.to()` before calling `to_ase()`.")

        batch_atoms = []
        add_msg = []
        names = set()

        batch = getattr(self, "batch")
        n_batches = batch.max() + 1
        for batch_idx in range(n_batches):
            data = self.get_example(batch_idx)
            pbc = getattr(data, "pbc", None)
            atomic_nums = getattr(data, "z", None)
            pos = getattr(data, "pos", None)
            cell = getattr(data, "cell", None)
            edge_index = getattr(data, "edge_index", None)

            energy = getattr(data, "energy", None)
            energies = getattr(data, "energies", None)
            forces = getattr(data, "forces", None)
            stress = getattr(data, "stress", None)

            if cell is not None:

                lattice = Lattice(matrix=torch.squeeze(cell).cpu().numpy(),
                                  pbc=pbc.view(-1).cpu().numpy())
                mol = Structure(
                    lattice=lattice,
                    species=atomic_nums.cpu().numpy().ravel(),
                    coords=pos.cpu().numpy(),
                    coords_are_cartesian=True)
            else:
                mol = Molecule(
                    species=atomic_nums.cpu().numpy().ravel(),
                    coords=pos.cpu().numpy(), )

            fields = {}
            if energies is not None:
                fields["energies"] = energies.cpu().numpy()
            if energy is not None:
                fields["energy"] = energy.cpu().numpy()
            if forces is not None:
                fields["forces"] = forces.cpu().numpy()
            if stress is not None:
                stress = stress.cpu().numpy()
                if self["stress"].shape[1] == 3:
                    fields["stress"] = full_3x3_to_voigt_6_stress(stress)
                else:
                    fields["stress"] = stress

            # add other information
            for key in extra_fields:
                if key == "edge_index":
                    fields[key] = edge_index.cpu().numpy().astype(np.long)
                else:
                    raise RuntimeError(
                        f"Extra field `{key}` isn't registered as node/edge/graph"
                    )

            batch_atoms.append(mol)
            add_msg.append(fields)
            names = names | set(fields.keys())

        names = list(names)

        kw = {i: [mi[i] for mi in add_msg] for i in names}

        if batch is not None:
            return batch_atoms, kw
        else:
            assert len(batch_atoms) == 1
            return batch_atoms[0], kw

    def to_atomicdatadict(self: Data, exclude_keys=tuple()) -> dict:
        """To dict to compare with AtomicDataDict"""
        keys = self.keys

        return {
            k: self[k]
            for k in keys
            if (k not in exclude_keys and self[k] is not None and isinstance(self[k], torch.Tensor))
        }

    def _per_static(self, fi, ana_mode, arr, inverse=False):
        if "per_atom" in ana_mode:
            natoms = getattr(self, "natoms")
            assert len(arr) == len(natoms), \
                f"Just features with same size of graph (sample number) could be divided.\n" \
                f"{fi} can't use 'per_atom' prefix."
            if inverse:
                arr = arr * natoms
            else:
                arr = arr / natoms
            res = {"per_atom": True}
        else:
            res = {"per_atom": False}
        return arr, res

    def statistics(
            self,
            fields: Union[str, List[str]],
            modes: Union[str, List[str]],
            unbiased: bool = True,
            by_dim=False,
    ) -> Dict[str, Dict[str, Union[float, torch.Tensor]]]:
        """
        statics the data.

        Args:
            fields: (str,list of str), names of property.
            modes: (str,list of str), names of in ["rms","unique","count","mean_std","max_min","self",
            "per_atom","per_atom",].
            unbiased: bool, unbiased for std.
            by_dim: (bool,list of bool) . statics according dim 0.

        Returns:
            data:dict
        """

        out = {}

        if not isinstance(fields, (list, tuple)):
            fields = [fields, ]

        if not isinstance(modes, (list, tuple)):
            modes = [modes, ]

        if not isinstance(by_dim, (list, tuple)):
            by_dim = [by_dim, ]

        if len(fields) == 1 and len(modes) > 1:
            fields = [fields[0]] * len(modes)
        elif len(fields) > 1 and len(modes) == 1:
            modes = [modes[0]] * len(fields)

        elif len(by_dim) == 1:
            by_dim = [by_dim[0]] * len(fields)

        assert len(fields) == len(modes) == len(by_dim)

        def up_add(fi, name, value):
            if fi in out:
                out[fi].update({name: value})
            else:
                out[fi] = {name: value}

        for fi, ana_mode, by_di in zip(fields, modes, by_dim):
            if not hasattr(self, fi):
                raise KeyError(f"No key named {fi}.")
            else:
                arr = getattr(self, fi)

                arr, perm = self._per_static(fi, ana_mode, arr, inverse=False)
                for k, v in perm.items():
                    up_add(fi, k, v)

                # compute statistics
                if "unique" in ana_mode:
                    # count integers
                    _, counts = torch.unique(torch.flatten(arr), return_counts=True, sorted=True)
                    up_add(fi, "unique", counts)

                if "count" in ana_mode:
                    # count integers
                    counts = arr.shape[0]
                    up_add(fi, "count", counts)

                elif "rms" in ana_mode:
                    # root-mean-square
                    if by_di:
                        counts = torch.sqrt(torch.mean(arr * arr, dim=0)).detach()
                    else:
                        counts = float(torch.sqrt(torch.mean(arr * arr)))
                    up_add(fi, "rms", counts)

                elif "mean_std" in ana_mode:
                    # mean and std
                    if by_di:
                        mean = torch.mean(arr, dim=0).detach()
                        std = torch.std(arr, dim=0, unbiased=unbiased).detach()
                    else:
                        mean = float(torch.mean(arr))
                        std = float(torch.std(arr, unbiased=unbiased))
                    up_add(fi, "mean", mean)
                    up_add(fi, "std", std)

                elif "max_min" in ana_mode:
                    if by_di:
                        maxs = torch.max(arr, dim=0).detach()
                        mins = torch.min(arr, dim=0).detach()
                    else:
                        maxs = float(torch.max(arr))
                        mins = float(torch.min(arr))
                    up_add(fi, "max", maxs)
                    up_add(fi, "min", mins)

                elif ana_mode is None or "self" in ana_mode:
                    up_add(fi, "self", arr)
                else:
                    raise NotImplementedError(f"Cannot handle statistics mode {ana_mode}")

        return out

    def scale(self, dct: Dict[str, Dict[str, Union[float, torch.Tensor]]]):
        """Scale the properties by the output of statistics."""
        for fi, ana_mode_dict in dct.items():
            if not hasattr(self, fi):
                pass
            else:
                arr = getattr(self, fi)

                ana_mode_name = "".join([i for i in ana_mode_dict.keys() if ana_mode_dict[i] is not False ])
                arr, _ = self._per_static(fi, ana_mode_name, arr, inverse=False)

                if "mean" in ana_mode_dict and "std" in ana_mode_dict:
                    arr = (arr - ana_mode_dict["mean"]) / ana_mode_dict["std"]
                elif "max" in ana_mode_dict and "min" in ana_mode_dict:
                    sc = ana_mode_dict["max"] - ana_mode_dict["min"]
                    arr = (arr - ana_mode_dict["min"]) / sc
                else:
                    pass
                setattr(self, fi, arr)
        return self

    def unscale(self, dct: Dict[str, Dict[str, Union[float, torch.Tensor]]]):
        """Un-scale the properties by the output of statistics."""
        for fi, ana_mode_dict in dct.items():
            if not hasattr(self, fi):
                pass
            else:
                arr = getattr(self, fi)

                if "mean" in ana_mode_dict and "std" in ana_mode_dict:
                    arr = (arr * ana_mode_dict["std"]) + ana_mode_dict["mean"]
                elif "max" in ana_mode_dict and "min" in ana_mode_dict:
                    sc = ana_mode_dict["max"] - ana_mode_dict["min"]
                    arr = arr * sc + ana_mode_dict["min"]
                else:
                    pass
                ana_mode_name = "".join([i for i in ana_mode_dict.keys() if ana_mode_dict[i] is not False ])
                arr, _ = self._per_static(fi, ana_mode_name, arr, inverse=True)

                setattr(self, fi, arr)

        return self

    def filter_index(self, prop: str, condition: Callable, independence: bool):
        """Filter by condition and return index."""
        batch = getattr(self, "batch")
        if independence:
            index = torch.Tensor([condition(getattr(i, prop)) for i in range(batch.max() + 1)])
        else:
            index = torch.Tensor(condition(getattr(self, prop)))
            assert len(index) == self.num_graphs, "only index with size == num_graphs could be used."
        return index

    def filter(self, prop: str, condition: Callable, independence: bool):
        """Filter by condition and return new."""
        index = self.filter_index(prop=prop, condition=condition, independence=independence)
        return self[index]

    def plot_show(self, fields: Union[str, List[str]] = None,
                  modes: Union[str, List[str]] = None,
                  by_dim: Union[bool, List[bool]] = None,
                  unbiased: bool = False, ):
        """plot

        Args:
            fields: (str,list of str), names of property.
            modes: (str,list of str), names of in ["rms","unique","count","mean_std","max_min","self",
            "per_atom","per_atom",].
            unbiased: bool, unbiased for std.
            by_dim: (bool,list of bool) . statics according dim 0.
        """
        if by_dim is None:
            by_dim = [False, True, False]
        if fields is None:
            fields = ["energy", "forces", "stress"]
        if modes is None:
            modes = ["self", "mean_std", "max_min"]

        import numpy as np

        import matplotlib.pyplot as plt

        dct = self.statistics(fields=fields, modes=modes, unbiased=unbiased, by_dim=by_dim)
        sp = {}
        sp1 = {}
        sp2 = {}
        for k, v in dct.items():
            for ki, vi in v.items():
                if isinstance(vi, (int, float)) or (isinstance(vi, torch.Tensor) and vi.ndim == 0):
                    sp.update({f"{k} ({ki})": np.array(vi)})
                elif isinstance(vi, torch.Tensor) and vi.ndim == 1:
                    sp1.update({f"{k} ({ki})": vi.view(-1).numpy()})
                elif isinstance(vi, torch.Tensor) and vi.ndim == 2:
                    sp2.update({f"{k} ({ki})": vi.numpy()})

                else:
                    print("Drop" f"{k}-{ki} due to large dim of tensor or error type.")
        start = 1 if len(sp) > 0 else 0
        shape = start + len(sp1) + len(sp2)

        plt.figure(figsize=(max((2.4 * len(sp1), 4.8)), 2 * shape))
        plt.subplot(shape, 1, 1, )
        plt.subplots_adjust(hspace=1)
        y = np.array(list(sp.values()))

        p1 = plt.bar(np.array(list(sp.keys())), y)
        plt.bar_label(p1, labels=y, fmt="%.2f", label_type='edge', padding=0, )
        plt.title("single properties")

        p = start
        for k, v in sp1.items():
            p = p + 1
            x_ = np.arange(len(v))
            plt.subplot(shape, 1, p)
            plt.xticks(x_, x_)
            plt.scatter(x_, v, marker="o" )
            plt.title(k)

        for k, v in sp2.items():
            p = p + 1
            plt.subplot(shape, 1, p)
            plt.imshow(v)
            plt.title(k)
        plt.savefig("data_range.jpg")
        try:
            plt.show()
        except BaseException:
            pass

    def plot_show_simple(self, fields: Union[str, List[str]] = "energy",
                         modes: Union[str, List[str]] = "self",
                         by_dim=False,
                         unbiased: bool = False, ):
        """quick plot energy.
        """
        self.plot_show(fields=fields, modes=modes, unbiased=unbiased, by_dim=by_dim)


def to_atomicdatadict(data: [Batch, Data, MtBatchData], exclude_keys=tuple()) -> dict:
    """quick to dict."""
    keys = data.keys
    return {k: data[k]
            for k in keys
            if (k not in exclude_keys and data[k] is not None and isinstance(data[k], torch.Tensor))
            }


def filter_upper(arr: torch.Tensor, theshold: float):
    return arr > theshold


def filter_lower(arr: torch.Tensor, theshold: float):
    return arr < theshold


def filter_equal(arr: torch.Tensor, theshold: float):
    return arr == theshold



import itertools
from collections.abc import Iterable
from functools import lru_cache

import numpy as np

from sympy import Matrix, symarray, solve
from sympy.core.numbers import One

one = One()


def spilt_tri_prism_z(data, six_index_xy, z_range=(0, 1), index_percent=True):
    """
    Spilt tri_prism in block data with six point, and the tri_prism is parallel to to Z axis.

    Thus use Three points of (x,y) and z_range to site the six point.


    Parameters
    ----------
    data: np.ndarray, tuple
        The matrix data self or the shape of matrix data.
    six_index_xy: np.ndarray
        Three x,y site of the six point of Tri-Prism.
        Examples:
            np.array([(0, 0.5), (0, 0), (0.7, 0.5)]) #if not index_percent
            np.array([(0, 5), (0, 0), (7, 5)]) #if not index_percent
    z_range:tuple
        z axis range.
        Examples:
            z_range=(0, 1) #if index_percent
            z_range=(0, 120) #if not index_percent
    index_percent:bool
        Treat the six_index_xy and z_range as percent of aixs or true index.

    Returns
    -------
    np.ndarray

    """
    if index_percent is False:
        if z_range[1] <= 1:
            raise NotImplemented("The z_range must be the shape of z axis without index_percent, such as (0,120)")
    six_index = []
    for i in six_index_xy:
        k = list(i)
        k.append(z_range[0])
        six_index.append(k)
        k = list(i)
        k.append(z_range[1])
        six_index.append(k)
    six_index = np.array(six_index)
    return spilt_tri_prism(data, six_index, index_percent)


def spilt_tri_prism(data, six_index, index_percent=True):
    """
    Spilt tri_prism in block data with six point.

    Parameters
    ----------
    data: np.ndarray, tuple
        The matrix data self or the shape of matrix data.
    six_index:np.ndarray
        Three x,y site of the six point of Tri-Prism.
        Examples:
            np.array([(0, 0.5), (0, 0), (0.7, 0.5)]) #if not index_percent
            np.array([(0, 5), (0, 0), (7, 5)]) #if not index_percent

    index_percent:bool
        Treat the six_index_xy and z_range as percent of aixs or true index.

    Returns
    -------
    np.ndarray
        Bool type with data' shape.
    """

    data_shape = data.shape if isinstance(data, np.ndarray) else data

    if not index_percent:
        six_index = six_index / data_shape

    assert np.max(six_index) <= 1
    assert np.min(six_index) >= 0

    # get the center of tri_prism
    center = np.mean(six_index, axis=0)
    center1 = np.append(center, 1)

    # get the five face of tri_prism
    si = tuple([tuple(i) for i in list(six_index)])
    face = get_face(si)

    # check the inner direction (to face)
    face_bool = np.sum(face * center1, axis=1)

    # check the index's direction  (to face)
    dims = data_shape
    coords = np.meshgrid(np.linspace(0, 1, dims[0]), np.linspace(0, 1, dims[1]), np.linspace(0, 1, dims[2]),
                         indexing="ij")
    xy_coords = np.vstack([coords[0].ravel(), coords[1].ravel(), coords[2].ravel(), np.ones_like(coords[2]).ravel()])

    corrds_bool = np.dot(face, xy_coords)

    # get the index's is in inner (to face)
    inface_index = corrds_bool * face_bool.reshape(-1, 1)
    inface_bool = inface_index >= 0

    # reback the index's to block
    inface_index_allTrue = np.all(inface_bool, axis=0)
    mark = inface_index_allTrue.reshape(dims, order="A")
    mark = mark.astype(int) # bool to 01
    return mark


@lru_cache(maxsize=300)
def get_face(six_index, tol=1e-4):
    """Get the five face of tri_prism."""
    six_index = np.array(six_index)
    all_com = []
    x = symarray('x', 4)

    # get the face fo all 3 point.
    for i in itertools.combinations(six_index, 3):
        a = np.array(i)
        a = np.concatenate((a, np.array([[1], [1], [1]])), axis=1)
        a = Matrix(a)

        z = a * x
        z = [sum(zi) for zi in z]

        re = solve(z, *x)

        if isinstance(re, dict):
            dd = {}
            for i in x:
                if i in re.keys():
                    dd[i] = re[i]
                else:
                    dd[i] = one
            all_com.append(dd)

    def sub_s(dic):
        """The intercept is reduced by 1, and the coefs are multiple to intercept"""
        coef = []
        for v in dic.values():
            if hasattr(v, 'free_symbols') and isinstance(v.free_symbols, Iterable):
                for free in v.free_symbols:
                    df = dic[free]
                    v = v.xreplace({free: df})
                coef.append(v)
            else:
                coef.append(v)
        return tuple(coef)

    coefs = []
    [coefs.append(sub_s(i)) for i in all_com]
    coefs = list(set(coefs))
    coefs.sort()
    six_index1 = np.concatenate((six_index, np.array([1, 1, 1, 1, 1, 1]).reshape(-1, 1)), axis=1)

    # delete the inner face, by judge the six points in only one side of face.
    sea = []
    for i in coefs:
        sums = np.sum(six_index1 * np.array(i), axis=1)
        th1 = np.all(sums >= -tol)
        th2 = np.all(sums <= tol)
        if th1:
            sea.append(i)
        elif th2:
            sea.append(i)
        else:
            pass

    sea = np.array(sea).astype(float) #sympy to float
    if sea.shape != (5, 4):
        print("The log")
        for n, i in enumerate(coefs):
            print("Face", i)
            sums = np.sum(six_index1 * np.array(i), axis=1)
            print("distinguish", sums)
            th1 = np.all(sums >= -tol)
            th2 = np.all(sums <= tol)
            if th1 or th2:
                print("number", n, "True ")
            else:
                print("number", n, "False ")
        raise TypeError("please check the log")

    return sea


def relative_location(points):
    """

    Parameters
    ----------
    points:
        np.ndarray
        shape(n_sample,n_dim)

    Returns
    -------
    new_points:
        np.ndarray
        shape(n_sample,n_dim)
    """
    maxs = np.max(points, axis=0)
    mins = np.min(points, axis=0)
    points = (points - mins) / (maxs - mins)
    return points


if __name__ == "__main__":
    # ss = spilt_tri_prism_z(np.ones([11, 12, 13]), np.array([(0, 5), (0, 0), (7, 5)]), z_range=(0, 8),
    #                        index_percent=False)
    #
    # ss2 = spilt_tri_prism_z(np.ones([11, 11, 13]), np.array([(0, 5), (0, 0), (7, 5)]), z_range=(0, 8),
    #                         index_percent=False)
    #
    # ss2 = spilt_tri_prism_z(np.ones([11, 12, 13]), np.array([(0, 5), (0, 0), (7, 5)]), z_range=(0, 8),
    #                         index_percent=False)

    a = np.array([5, 2.0001, 2, 4.0004, 4, 1])
    b = np.sort(a)
    c = np.diff(b)
    c=np.append(c,np.inf)
    d = np.where(abs(c)>1e-3)
    res=b[d]


"""
Examples for ase db:
    >>> from hpc2ml.data.structuretodata import StructureToData
    >>> addsap = StructureToData()
    >>> source_path = "."
    >>> res = addsap.sparse_source_data(source_file="data.db", source_path=source_path, fmt="ase",)

Examples for Vasp:
    >>> from hpc2ml.data.structuretodata import StructureToData
    >>> addsap = StructureToData()
    >>> source_path = find_leaf_path("./data") # get path list
    >>> res = addsap.sparse_source_data(source_file="vasprun.xml",
    ... source_path=source_path, fmt="vasprun",)
    >>> res2 = addsap.sparse_source_data(source_file="vasprun.xml",
    ... source_path=source_path, fmt="vasprun_traj",space=5)

Examples for csv:
    >>> from hpc2ml.data.structuretodata import StructureToData
    >>> addsap = StructureToData()
    >>> source_path = "."
    >>> res = addsap.sparse_source_data(source_file="tb.csv", source_path=source_path, fmt="csv",)

"""

import os
from typing import Union, Sequence

from mgetool.tool import parallelize


def sparse_source_data(source_file: str = "vasprun.xml", source_path: Union[Sequence, str] = ".",
                       fmt: str = "vasprun",
                       n_jobs=4, **kwargs):
    """
    Sparse data by different function.

    Args:
        source_file: file name to sparse.
        source_path: (str,list), path or paths.
        fmt: str, load function named  "sparse_{fmt}" in from ``hpc2ml.data.io`` .
        n_jobs: int, the parallel number to load,
        **kwargs: dict, the parameter in "sparse_{fmt}".

    Returns:
        data_dict:dict, data
    """
    if fmt == "auto":
        suffix = source_file.split(".")[-1]
        fmt = {"xml": "vasprun", "db": "ase", "csv": "csv"}[suffix]

    if isinstance(source_path, str):
        source_path = [source_path, ]

    source_file = [os.path.join(pathi, source_file) for pathi in source_path]
    from hpc2ml.data import io
    func = getattr(io, f"sparse_{fmt}")

    def func2(i):
        try:
            dicti = func(i, **kwargs)
        except:
            dicti = {}
            print(f"Error for : {i}")
        return dicti

    dct = {}
    res = parallelize(n_jobs=n_jobs, func=func2, iterable=source_file, tq=True, desc="sparse the source data")
    [dct.update(i) for i in res]

    # for i in source_file:
    #     try:
    #         dcti = func(i)
    #         dct.update(dcti)
    #     except:
    #         print(f"Error: {i}.")

    return dct



import os
import shutil
from pathlib import PurePath
from typing import Union, Tuple, List

from ase.db import connect
from ase.db.core import Database
from ase.db.row import AtomsRow

from .ext_row import check_name_tup, atoms_row_rename


def check_postfix(name: Union[Database, str, dict]):
    if isinstance(name, PurePath):
        name = str(name)

    if isinstance(name, dict):
        postfix = '.json'
    elif isinstance(name, str):
        postfix = os.path.splitext(name)[-1]
    else:
        from ase.db.jsondb import JSONDatabase
        from ase.db.sqlite import SQLite3Database

        if isinstance(name, JSONDatabase):
            postfix = '.json'
        elif isinstance(name, SQLite3Database):
            postfix = '.db'
        else:
            postfix = ''

    return postfix


def db_append(database1: Database, database2: Database, parallel=False) -> None:
    """Replace the in situ."""
    if parallel:
        with database1:
            with database2:
                for ar in database2.select():
                    id = database1.reserve(name=str(ar))
                    if id is None:
                        continue
                    database1.write(ar, ar.key_value_pairs, ar.data, id=id)
    else:
        with database1:
            with database2:
                for ar in database2.select():
                    database1.write(ar, ar.key_value_pairs, ar.data)


def db_remove(database: Database, sli: [slice, List, Tuple]) -> None:
    """Delete the in situ!!!!!,Keep the id in sli!!!."""
    if isinstance(sli, Tuple):
        assert len(sli) <= 3, 'just accept (start,stop,[step])'
        sli = slice(*sli)
    if isinstance(sli, slice):
        assert sli.start >= 1, "slice for database start from 1"
        if sli.step is None:
            sli = slice(sli.start, sli.stop, 1)
        sli_list = list(range(sli.start, sli.stop, sli.step))
    else:
        sli_list = sli
    with database:
        database.delete(sli_list)
    print(sli_list, "is deleted")


def db_slice(database: Database, sli: [slice, Tuple], new_database_name: str = None, paraller=False) -> Union[
    Database, List[AtomsRow]]:
    """Get slice of database, in new database, Keep the id in sli!!!"""
    if isinstance(database, str):
        database = connect(database)

    if isinstance(sli, Tuple):
        assert len(sli) <= 3, 'just accept (start,stop,[step])'
        sli = slice(*sli)

    assert isinstance(sli, slice) and sli.start >= 1, "slice for database start from 1"
    if sli.step is None:
        sli = slice(sli.start, sli.stop, 1)
    if isinstance(new_database_name, str):
        new_database_name = os.path.splitext(new_database_name)[0]
        new_database_name = "".join((new_database_name, "{}"))
        database1_ = database.__class__(new_database_name.format(check_postfix(database)))
        if paraller:
            with database:
                with database1_:
                    for i in range(sli.start, sli.stop, sli.step):
                        id = database1_.reserve(name=i)
                        if id is None:
                            continue
                        ar = database[i]
                        database1_.write(ar, ar.key_value_pairs, ar.data, id=id)
        else:
            with database:
                with database1_:
                    for i in range(sli.start, sli.stop, sli.step):
                        ar = database[i]
                        database1_.write(ar, ar.key_value_pairs, ar.data)
    else:
        database1_ = []
        with database:
            for i in range(sli.start, sli.stop, sli.step):
                database1_.append(database[i])
    return database1_


def db_cat(database1: Union[Database, str], database2: Union[Database, str], new_database_name: str) -> Database:
    """new database, in new database."""
    new_database_name = os.path.splitext(new_database_name)[0]
    new_database_name = "".join((new_database_name, "{}"))

    if isinstance(database1, Database):
        database1 = database1.filename

    if isinstance(database1, str):
        if isinstance(database2, str):
            database2 = connect(database2)

        assert os.path.isfile(database1)
        new = new_database_name.format(check_postfix(database1))
        shutil.copy(database1, new)
        new = connect(new)
        db_append(new, database2)
        return new

    else:
        # old deprecated
        if isinstance(database1, str):
            database1 = connect(database1)
        if isinstance(database2, str):
            database2 = connect(database2)
        database1_ = database1.__class__(new_database_name.format(check_postfix(database1)))

        with database1_:
            with database1:
                for ar in database1.select(selection=None):
                    database1_.write(ar, ar.key_value_pairs, ar.data)
            with database2:
                for ar in database2.select(selection=None):
                    database1_.write(ar, ar.key_value_pairs, ar.data)
        return database1_


def db_ids(database: Database, selection=None, **kwargs) -> Tuple:
    """Get ids.
    """
    ids = []
    with database:
        for row in database.select(selection, **kwargs):
            ids.append(row.id)

    t = ids[0]
    i = ids[0]
    pstr = "({}-".format(t)
    for i in ids:
        if i - t <= 1:
            pass
        else:
            pstr = pstr + "{},{}-".format(t, i)
        t = i
    pstr = pstr + "{})".format(i)
    print(pstr)
    return tuple(ids)


def db_rename(database: Database, name_pair=(("old_name1", "new_name1"),), check=True, selection=None,
              **kwargs) -> Database:
    if check:
        check_name_tup(name_pair)

    if os.path.splitext(database.filename)[1] == '.json':
        js = True
    else:
        js = False
    with database:
        for row in database.select(selection, **kwargs):
            new_row = atoms_row_rename(row, name_pair=name_pair, check=False)
            kvp = new_row.key_value_pairs
            data = new_row.get('data', {})

            if js:
                database._write(new_row, kvp, data, row.id)
            else:
                database._update(row.id, kvp, data)

    return database


def db_transform(database, new_database_name="new_database.json"):
    """Convert ase.json to ase.db, or inverse."""

    assert ".db" in new_database_name or ".json" in new_database_name
    if os.path.isfile(new_database_name):
        raise FileExistsError(new_database_name, "is exist, please change the `new_database_name`.")

    to_type = os.path.splitext(new_database_name)[1]

    assert to_type in [".json", ".db"]

    from ase.db.jsondb import JSONDatabase
    from ase.db.sqlite import SQLite3Database

    if isinstance(database, str):
        database = connect(database)
    database1_ = connect(new_database_name)

    if isinstance(database, JSONDatabase) and to_type == ".json":
        print("Initial type is json, do nothing.")
        return
    elif isinstance(database, SQLite3Database) and to_type == ".db":
        print("Initial type is db (SQLite3), do nothing.")
        return

    else:
        with database:
            with database1_:
                for ar in database.select():
                    database1_.write(ar, ar.key_value_pairs, ar.data)
        return database1_



"""Pure torch functions."""

import math

import torch
from torch_geometric.nn import radius_graph
# from torch_cluster import radius_graph
from torch_geometric.utils import remove_self_loops
from torch_scatter import segment_coo, segment_csr


def add_edge_total(data, use_pbc=True, use_exist_index=True, **kwargs):
    """Total function to add edge_weight,edge_attr,data.edge_index."""
    if not use_pbc:
        if hasattr(data, "edge_index") and use_exist_index:
            data = add_edge_no_pbc_from_index(data, **kwargs)
        else:
            data = add_edge_no_pbc(data, **kwargs)
    else:
        if hasattr(data, "edge_index") and use_exist_index:
            data = add_edge_pbc_from_index(data, **kwargs)
        else:
            data = add_edge_pbc(data, **kwargs)
    return data


def add_edge_no_pbc(data, cutoff=5.0):
    """Add edge_weight,edge_attr,data.edge_index ignore pbc."""
    edge_index = radius_graph(data.pos, r=cutoff, batch=data.batch, flow='source_to_target')
    edge_index, _ = remove_self_loops(edge_index)
    row, col = edge_index[0], edge_index[1]

    distance_vectors = data.pos[row] - data.pos[col]

    edge_weight = distance_vectors.norm(dim=-1)

    data.edge_attr = distance_vectors
    data.edge_index = edge_index
    data.edge_weight = edge_weight
    return data


def add_edge_no_pbc_from_index(data, dmin=0.0, dmax=6.0):
    """Add edge_weight,edge_attr ignore pbc from the old edge_index."""

    edge_index = data.edge_index

    row, col = edge_index[0], edge_index[1]

    distance_vectors = data.pos[row] - data.pos[col]

    edge_weight = distance_vectors.norm(dim=-1)

    data.edge_index = edge_index
    data.edge_weight = edge_weight
    data.edge_attr = distance_vectors

    return data


def compute_neighbors(data):
    """Add neighbors for each atom"""
    # Get number of neighbors
    # segment_coo assumes sorted index
    edge_index = data.edge_index
    ones = edge_index[1].new_ones(1).expand_as(edge_index[1])
    num_neighbors = segment_coo(
        ones, edge_index[1], dim_size=data.natoms.sum()
    )

    # Get number of neighbors per image
    image_indptr = torch.zeros(
        data.natoms.shape[0] + 1, device=data.pos.device, dtype=torch.long
    )
    image_indptr[1:] = torch.cumsum(data.natoms, dim=0)
    neighbors = segment_csr(num_neighbors, image_indptr)
    data.neighbors = neighbors
    return data


def add_edge_pbc_from_index(data, **kwargs):
    """Add edge_weight,edge_attr with pbc from the old edge_index."""

    if not hasattr(data, "neighbors"):
        data = compute_neighbors(data)
    cell = data.cell
    edge_index = data.edge_index
    neighbors = data.neighbors
    pos = data.pos
    cell_offsets = data.cell_offsets

    edge_index, edge_weight, cell_offsets, distance_vectors = _add_edge_pbc_from_index(pos,
                                                                                       edge_index,
                                                                                       cell,
                                                                                       cell_offsets,
                                                                                       neighbors,
                                                                                       )

    data.cell_offsets = cell_offsets
    data.edge_index = edge_index
    data.edge_weight = edge_weight
    data.edge_attr = distance_vectors

    return data


def add_edge_pbc(data, cutoff=5.0, max_neighbors=16):
    """Add edge_weight,edge_attr with pbc."""
    edge_index, cell_offsets, neighbors = _radius_graph_pbc(
        data, cutoff, max_neighbors
    )

    edge_index, edge_weight, cell_offsets, distance_vectors = _add_edge_pbc_from_index(
        data.pos,
        edge_index,
        data.cell,
        cell_offsets,
        neighbors,
    )

    data.edge_attr = distance_vectors
    data.cell_offsets = cell_offsets
    data.edge_index = edge_index
    data.edge_weight = edge_weight

    return data


def _add_edge_pbc_from_index(pos,
                             edge_index,
                             cell,
                             cell_offsets,
                             neighbors,
                             ):
    device = cell.device
    row, col = edge_index
    distance_vectors = pos[row] - pos[col]
    neighbors = neighbors.to(device)
    cell = torch.repeat_interleave(cell, neighbors, dim=0)
    offsets = cell_offsets.float().view(-1, 1, 3).bmm(cell.float()).view(-1, 3)
    distance_vectors += offsets

    # compute distances
    distances = distance_vectors.norm(dim=-1)

    # redundancy: remove zero distances
    nonzero_idx = torch.arange(len(distances), device=device)[distances != 0]

    edge_index = edge_index[:, nonzero_idx]
    edge_weight = distances[nonzero_idx]
    cell_offsets = cell_offsets[nonzero_idx]

    return edge_index, edge_weight, cell_offsets, distance_vectors


def _get_max_neighbors_mask(
        natoms, index, atom_distance, max_num_neighbors_threshold
):
    """
    Give a mask that filters out edges so that each atom has at most
    `max_num_neighbors_threshold` neighbors.
    Assumes that `index` is sorted.
    """
    device = natoms.device
    num_atoms = natoms.sum()

    # Get number of neighbors
    # segment_coo assumes sorted index
    ones = index.new_ones(1).expand_as(index)
    num_neighbors = segment_coo(ones, index, dim_size=num_atoms)
    max_num_neighbors = num_neighbors.max()
    num_neighbors_thresholded = num_neighbors.clamp(
        max=max_num_neighbors_threshold
    )

    # Get number of (thresholded) neighbors per image
    image_indptr = torch.zeros(
        natoms.shape[0] + 1, device=device, dtype=torch.long
    )
    image_indptr[1:] = torch.cumsum(natoms, dim=0)
    num_neighbors_image = segment_csr(num_neighbors_thresholded, image_indptr)

    # If max_num_neighbors is below the threshold, return early
    if (
            max_num_neighbors <= max_num_neighbors_threshold
            or max_num_neighbors_threshold <= 0
    ):
        mask_num_neighbors = torch.tensor(
            [True], dtype=bool, device=device
        ).expand_as(index)
        return mask_num_neighbors, num_neighbors_image

    # Create a tensor of size [num_atoms, max_num_neighbors] to sort the distances of the neighbors.
    # Fill with infinity so we can easily remove unused distances later.
    distance_sort = torch.full(
        [num_atoms * max_num_neighbors], torch.inf, device=device
    )

    # Create an index map to map distances from atom_distance to distance_sort
    # index_sort_map assumes index to be sorted
    index_neighbor_offset = torch.cumsum(num_neighbors, dim=0) - num_neighbors
    index_neighbor_offset_expand = torch.repeat_interleave(
        index_neighbor_offset, num_neighbors
    )
    index_sort_map = (
            index * max_num_neighbors
            + torch.arange(len(index), device=device)
            - index_neighbor_offset_expand
    )
    distance_sort.index_copy_(0, index_sort_map, atom_distance)
    distance_sort = distance_sort.view(num_atoms, max_num_neighbors)

    # Sort neighboring atoms based on distance
    distance_sort, index_sort = torch.sort(distance_sort, dim=1)
    # Select the max_num_neighbors_threshold neighbors that are closest
    distance_sort = distance_sort[:, :max_num_neighbors_threshold]
    index_sort = index_sort[:, :max_num_neighbors_threshold]

    # Offset index_sort so that it indexes into index
    index_sort = index_sort + index_neighbor_offset.view(-1, 1).expand(
        -1, max_num_neighbors_threshold
    )
    # Remove "unused pairs" with infinite distances
    mask_finite = torch.isfinite(distance_sort)
    index_sort = torch.masked_select(index_sort, mask_finite)

    # At this point index_sort contains the index into index of the
    # closest max_num_neighbors_threshold neighbors per atom
    # Create a mask to remove all pairs not in index_sort
    mask_num_neighbors = torch.zeros(len(index), device=device, dtype=bool)
    mask_num_neighbors.index_fill_(0, index_sort, True)

    return mask_num_neighbors, num_neighbors_image


def _radius_graph_pbc(data, radius, max_num_neighbors_threshold, pbc=None):
    if pbc is None:
        pbc = [True, True, True]
    device = data.pos.device
    batch_size = len(data.natoms)

    if hasattr(data, "pbc"):
        data.pbc = torch.atleast_2d(data.pbc)
        for i in range(3):
            if not torch.any(data.pbc[:, i]).item():
                pbc[i] = False
            elif torch.all(data.pbc[:, i]).item():
                pbc[i] = True
            else:
                raise RuntimeError(
                    "Different structures in the batch have different PBC configurations. This is not currently supported."
                )

    # position of the atoms
    atom_pos = data.pos

    # Before computing the pairwise distances between atoms, first create a list of atom indices to compare for the entire batch
    num_atoms_per_image = data.natoms
    num_atoms_per_image_sqr = (num_atoms_per_image ** 2).long()

    # index offset between images
    index_offset = (
            torch.cumsum(num_atoms_per_image, dim=0) - num_atoms_per_image
    )

    index_offset_expand = torch.repeat_interleave(
        index_offset, num_atoms_per_image_sqr
    )
    num_atoms_per_image_expand = torch.repeat_interleave(
        num_atoms_per_image, num_atoms_per_image_sqr
    )

    # Compute a tensor containing sequences of numbers that range from 0 to num_atoms_per_image_sqr for each image
    # that is used to compute indices for the pairs of atoms. This is a very convoluted way to implement
    # the following (but 10x faster since it removes the for loop)
    # for batch_idx in range(batch_size):
    #    batch_count = torch.cat([batch_count, torch.arange(num_atoms_per_image_sqr[batch_idx], device=device)], dim=0)
    num_atom_pairs = torch.sum(num_atoms_per_image_sqr)
    index_sqr_offset = (
            torch.cumsum(num_atoms_per_image_sqr, dim=0) - num_atoms_per_image_sqr
    )
    index_sqr_offset = torch.repeat_interleave(
        index_sqr_offset, num_atoms_per_image_sqr
    )
    atom_count_sqr = (
            torch.arange(num_atom_pairs, device=device) - index_sqr_offset
    )

    # Compute the indices for the pairs of atoms (using division and mod)
    # If the systems get too large this apporach could run into numerical precision issues
    index1 = (
                 torch.div(
                     atom_count_sqr, num_atoms_per_image_expand, rounding_mode="floor"
                 )
             ) + index_offset_expand
    index2 = (
                     atom_count_sqr % num_atoms_per_image_expand
             ) + index_offset_expand
    # Get the positions for each atom
    pos1 = torch.index_select(atom_pos, 0, index1)
    pos2 = torch.index_select(atom_pos, 0, index2)

    # Calculate required number of unit cells in each direction.
    # Smallest distance between planes separated by a1 is
    # 1 / ||(a2 x a3) / V||_2, since a2 x a3 is the area of the plane.
    # Note that the unit cell volume V = a1 * (a2 x a3) and that
    # (a2 x a3) / V is also the reciprocal primitive vector
    # (crystallographer's definition).

    cross_a2a3 = torch.cross(data.cell[:, 1], data.cell[:, 2], dim=-1)
    cell_vol = torch.sum(data.cell[:, 0] * cross_a2a3, dim=-1, keepdim=True)

    if pbc[0]:
        inv_min_dist_a1 = torch.norm(cross_a2a3 / cell_vol, p=2, dim=-1)
        rep_a1 = torch.ceil(radius * inv_min_dist_a1)
    else:
        rep_a1 = data.cell.new_zeros(1)

    if pbc[1]:
        cross_a3a1 = torch.cross(data.cell[:, 2], data.cell[:, 0], dim=-1)
        inv_min_dist_a2 = torch.norm(cross_a3a1 / cell_vol, p=2, dim=-1)
        rep_a2 = torch.ceil(radius * inv_min_dist_a2)
    else:
        rep_a2 = data.cell.new_zeros(1)

    if pbc[2]:
        cross_a1a2 = torch.cross(data.cell[:, 0], data.cell[:, 1], dim=-1)
        inv_min_dist_a3 = torch.norm(cross_a1a2 / cell_vol, p=2, dim=-1)
        rep_a3 = torch.ceil(radius * inv_min_dist_a3)
    else:
        rep_a3 = data.cell.new_zeros(1)

    # Take the max over all images for uniformity. This is essentially padding.
    # Note that this can significantly increase the number of computed distances
    # if the required repetitions are very different between images
    # (which they usually are). Changing this to sparse (scatter) operations
    # might be worth the effort if this function becomes a bottleneck.
    max_rep = [rep_a1.max(), rep_a2.max(), rep_a3.max()]

    # Tensor of unit cells
    cells_per_dim = [
        torch.arange(-rep, rep + 1, device=device, dtype=torch.float)
        for rep in max_rep
    ]
    unit_cell = torch.cartesian_prod(*cells_per_dim)
    num_cells = len(unit_cell)
    unit_cell_per_atom = unit_cell.view(1, num_cells, 3).repeat(
        len(index2), 1, 1
    )
    unit_cell = torch.transpose(unit_cell, 0, 1)
    unit_cell_batch = unit_cell.view(1, 3, num_cells).expand(
        batch_size, -1, -1
    )

    # Compute the x, y, z positional offsets for each cell in each image
    data_cell = torch.transpose(data.cell, 1, 2)
    pbc_offsets = torch.bmm(data_cell, unit_cell_batch)
    pbc_offsets_per_atom = torch.repeat_interleave(
        pbc_offsets, num_atoms_per_image_sqr, dim=0
    )

    # Expand the positions and indices for the 9 cells
    pos1 = pos1.view(-1, 3, 1).expand(-1, -1, num_cells)
    pos2 = pos2.view(-1, 3, 1).expand(-1, -1, num_cells)
    index1 = index1.view(-1, 1).repeat(1, num_cells).view(-1)
    index2 = index2.view(-1, 1).repeat(1, num_cells).view(-1)
    # Add the PBC offsets for the second atom
    pos2 = pos2 + pbc_offsets_per_atom

    # Compute the squared distance between atoms
    atom_distance_sqr = torch.sum((pos1 - pos2) ** 2, dim=1)
    atom_distance_sqr = atom_distance_sqr.view(-1)

    # Remove pairs that are too far apart
    mask_within_radius = torch.le(atom_distance_sqr, radius * radius)
    # Remove pairs with the same atoms (distance = 0.0)
    mask_not_same = torch.gt(atom_distance_sqr, 0.0001)
    mask = torch.logical_and(mask_within_radius, mask_not_same)
    index1 = torch.masked_select(index1, mask)
    index2 = torch.masked_select(index2, mask)
    unit_cell = torch.masked_select(
        unit_cell_per_atom.view(-1, 3), mask.view(-1, 1).expand(-1, 3)
    )
    unit_cell = unit_cell.view(-1, 3)
    atom_distance_sqr = torch.masked_select(atom_distance_sqr, mask)

    mask_num_neighbors, num_neighbors_image = _get_max_neighbors_mask(
        natoms=data.natoms,
        index=index1,
        atom_distance=atom_distance_sqr,
        max_num_neighbors_threshold=max_num_neighbors_threshold,
    )

    if not torch.all(mask_num_neighbors):
        # Mask out the atoms to ensure each atom has at most max_num_neighbors_threshold neighbors
        index1 = torch.masked_select(index1, mask_num_neighbors)
        index2 = torch.masked_select(index2, mask_num_neighbors)
        unit_cell = torch.masked_select(
            unit_cell.view(-1, 3), mask_num_neighbors.view(-1, 1).expand(-1, 3)
        )
        unit_cell = unit_cell.view(-1, 3)

    edge_index = torch.stack((index2, index1))

    return edge_index, unit_cell, num_neighbors_image


def distribute_edge(data, r_cs, r_c):
    """For data.edge_attr is the x,y,z site of atom."""
    assert hasattr(data, "edge_attr") and data.edge_attr is not None
    assert data.edge_attr.shape[1] == 3
    if not hasattr(data, "edge_weight") or data.edge_weight is None:
        data.edge_weight = torch.linalg.norm(data.edge_attr, dim=1)

    wei = data.edge_weight

    attr = torch.cat((wei.reshape(-1, 1), data.edge_attr), dim=1)

    sr = torch.clone(attr[:, 0])
    r_m1 = sr <= r_cs
    r_m3 = sr >= r_c
    r_m2 = ~(r_m3 | r_m1)

    sr[r_m1] = 1 / sr[r_m1]

    u = (sr[r_m2] - r_cs) / (r_c - r_cs)
    sr[r_m2] = 1 / sr[r_m2] * (0.5 * torch.cos(math.pi * u) + 0.5)  # from deep Potential
    sr[r_m3] = 0

    attr = attr / attr[:, 0].reshape(-1, 1) * sr.reshape(-1, 1)

    data.edge_attr = attr
    data.edge_weight = sr
    return data



# -*- coding: utf-8 -*-

import os
import random

# @Time  : 2023/1/24 17:01
# @Author : boliqq07
# @Software: PyCharm
# @License: MIT License
import numpy as np
import torch


def seed_torch(seed=0):
    random.seed(seed)
    os.environ['PYTHONHASHSEED'] = str(seed)  # 为了禁止hash随机化，使得实验可复现
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)  # if you are using multi-GPU.
    torch.backends.cudnn.benchmark = False
    torch.backends.cudnn.deterministic = True



import warnings
from typing import List, Union, Optional

import torch
from e3nn.o3 import Irreps
from e3nn.util.jit import compile_mode
from torch_geometric.data import Batch, Data

from hpc2ml.data.nequlip_support import AtomicDataDict
from hpc2ml.data.batchdata import to_atomicdatadict
from hpc2ml.data.nequlip_support.AtomicDataDict import ST
from ._graph_mixin import GraphModuleMixin

def voigt_6_to_full_3x3_stress(stress_vector):
    """
    Form a 3x3 stress matrix from a 6 component vector in Voigt notation
    """
    s1, s2, s3, s4, s5, s6 = torch.split(stress_vector, 6)
    return torch.cat([s1, s6, s5,s6, s2, s4,s5, s4, s3]).resize((3, 3))


class GradientOutput(GraphModuleMixin, torch.nn.Module):
    r"""Wrap a model and include as an output its gradient.

    Args:
        func: the model to wrap
        of: the name of the output in_field of ``func`` to take the gradient with respect to.
         The in_field must be a single scalar (i.e. have irreps ``0e``)
        wrt: the input in_field(s) of ``func`` to take the gradient of ``of`` with regards to.
        out_field: the in_field in which to return the computed gradients.
         Defaults to ``f"d({of})/d({wrt})"`` for each in_field in ``wrt``.
        sign: either 1 or -1; the returned gradient is multiplied by this.
    """
    sign: float
    _negate: bool
    skip: bool

    def __init__(
            self,
            func,
            of: str="energy",
            wrt: Union[str, List[str]] = "pos",
            out_field: Optional[List[str]] = None,
            sign: float = 1.0,
    ):
        super().__init__()
        sign = float(sign)
        assert sign in (1.0, -1.0)
        self.sign = sign
        self._negate = sign == -1.0
        self.of = of
        self.skip = False

        # TO DO: maybe better to forces using list?
        if isinstance(wrt, str):
            wrt = [wrt]
        if isinstance(out_field, str):
            out_field = [out_field]
        self.wrt = wrt
        self.func = func
        if out_field is None:
            self.out_field = [f"d({of})/d({e})" for e in self.wrt]
        else:
            assert len(out_field) == len(
                self.wrt
            ), "Out in_field names must be given for all w.r.t tensors"
            self.out_field = out_field

        # check and init irreps
        self._init_irreps(
            irreps_in=func.irreps_in,
            my_irreps_in={of: Irreps("0e")},
            irreps_out=func.irreps_out,
        )

        # The gradient of a single scalar w.r.t. something of a given shape and irrep just has that shape and irrep
        # Ex.: gradient of energy (0e) w.r.t. position vector (L=1) is also an L = 1 vector
        self.irreps_out.update(
            {f: self.irreps_in[wrt] for f, wrt in zip(self.out_field, self.wrt)}
        )

    def forward(self, data: Union[Batch, Data, dict]):

        if not isinstance(data, dict):
            data = to_atomicdatadict(data)

        if self.skip:
            return self.func(data)

        # set req grad
        wrt_tensors = []
        old_requires_grad: List[bool] = []
        for k in self.wrt:
            old_requires_grad.append(data[k].requires_grad)
            data[k].requires_grad_(True)
            wrt_tensors.append(data[k])
        # run func
        data = self.func(data)
        # Get grads
        grads = torch.autograd.grad(
            # TODO:
            # This makes sense for scalar batch-level or batch-wise outputs, specifically because d(sum(batches))/d wrt = sum(d batch / d wrt) = d my_batch / d wrt
            # for a well-behaved example level like energy where d other_batch / d wrt is always zero. (In other words, the energy of example 1 in the batch is completely unaffect by changes in the position of atoms in another example.)
            # This should work for any gradient of energy, but could act suspiciously and unexpectedly for arbitrary gradient outputs, if they ever come up
            [data[self.of].sum()],
            wrt_tensors,
            create_graph=self.training,  # needed to allow gradients of this output during training
        )
        # return
        # grad is optional[tensor]?
        for out, grad in zip(self.out_field, grads):
            if grad is None:
                # From the docs: "If an output doesn’t require_grad, then the gradient can be None"
                raise RuntimeError("Something is wrong, gradient couldn't be computed")

            if self._negate:
                grad = torch.neg(grad)
            data[out] = grad

        # unset requires_grad_
        for req_grad, k in zip(old_requires_grad, self.wrt):
            data[k].requires_grad_(req_grad)

        return data


class PartialForceOutput(GraphModuleMixin, torch.nn.Module):
    r"""Generate partial and total forces from an energy model.

    Args:
        func: the energy model
        vectorize: the vectorize option to ``torch.autograd.functional.jacobian``,
            false by default since it doesn't work well.
    """
    vectorize: bool

    def __init__(
            self,
            func,
            vectorize: bool = False,
            vectorize_warnings: bool = False,
    ):
        super().__init__()
        self.func = func
        self.vectorize = vectorize
        if vectorize_warnings:
            # See https://pytorch.org/docs/stable/generated/torch.autograd.functional.jacobian.html
            torch._C._debug_only_display_vmap_fallback_warnings(True)

        # check and init irreps
        self._init_irreps(
            irreps_in=func.irreps_in,
            my_irreps_in={AtomicDataDict.PER_ATOM_ENERGY_KEY: Irreps("0e")},
            irreps_out=func.irreps_out,
        )
        self.irreps_out[AtomicDataDict.PARTIAL_FORCE_KEY] = Irreps("1o")
        self.irreps_out[AtomicDataDict.FORCE_KEY] = Irreps("1o")

    def forward(self, data: Union[Batch, Data, dict]):

        if not isinstance(data, dict):
            data = to_atomicdatadict(data)

        data = data.copy()
        out_data = {}

        def wrapper(pos: torch.Tensor) -> torch.Tensor:
            """Wrapper from pos to atomic energy"""
            nonlocal data, out_data
            data[AtomicDataDict.POSITIONS_KEY] = pos
            out_data = self.func(data)
            return out_data[AtomicDataDict.PER_ATOM_ENERGY_KEY].squeeze(-1)

        pos = data[AtomicDataDict.POSITIONS_KEY]

        partial_forces = torch.autograd.functional.jacobian(
            func=wrapper,
            inputs=pos,
            create_graph=self.training,  # needed to allow gradients of this output during training
            vectorize=self.vectorize,
        )
        partial_forces = partial_forces.negative()
        # output is [n_at, n_at, 3]

        out_data[AtomicDataDict.PARTIAL_FORCE_KEY] = partial_forces
        out_data[AtomicDataDict.FORCE_KEY] = partial_forces.sum(dim=0)

        return out_data


class StressOutput(GraphModuleMixin, torch.nn.Module):
    r"""Compute stress (and forces) using autograd of an energy model.

    See:
        Knuth et. al. Comput. Phys. Commun 190, 33-50, 2015
        https://pure.mpg.de/rest/items/item_2085135_9/component/file_2156800/content

    Args:
        func: the energy model to wrap
        do_forces: whether to compute forces as well
    """
    do_forces: bool

    def __init__(
            self,
            func,
            do_forces: bool = True,
    ):
        super().__init__()

        # warnings.warn(
        #     "!! Stresses in NequIP are in BETA and UNDER DEVELOPMENT: _please_ carefully check the sanity of your results and report any (potential) issues on the GitHub"
        # )

        if not do_forces:
            raise NotImplementedError
        self.do_forces = do_forces

        self.func = func

        # check and init irreps
        self._init_irreps(
            irreps_in=self.func.irreps_in.copy(),
            irreps_out=self.func.irreps_out.copy(),
        )
        self.irreps_out[AtomicDataDict.FORCE_KEY] = "1o"
        self.irreps_out[AtomicDataDict.STRESS_KEY] = "3x1o"
        self.irreps_out[AtomicDataDict.VIRIAL_KEY] = "3x1o"

        # for torchscript compat
        self.register_buffer("_empty", torch.Tensor())

    def forward(self, data: Union[Batch, Data, dict]):

        if not isinstance(data, dict):
            data = to_atomicdatadict(data)

        data = AtomicDataDict.with_batch(data)

        batch = data[AtomicDataDict.BATCH_KEY]
        num_batch: int = int(batch.max().cpu().item()) + 1
        pos = data[AtomicDataDict.POSITIONS_KEY]

        has_cell: bool = AtomicDataDict.CELL_KEY in data

        if has_cell:
            orig_cell = data[AtomicDataDict.CELL_KEY]
            # Make the cell per-batch
            cell = orig_cell.view(-1, 3, 3).expand(num_batch, 3, 3)
            data[AtomicDataDict.CELL_KEY] = cell
        else:
            # torchscript
            orig_cell = self._empty
            cell = self._empty
        # Add the displacements
        # the GradientOutput will make them require grad
        # See SchNetPack code:
        # https://github.com/atomistic-machine-learning/schnetpack/blob/master/src/schnetpack/atomistic/model.py#L45
        # SchNetPack issue:
        # https://github.com/atomistic-machine-learning/schnetpack/issues/165
        # Paper they worked from:
        # Knuth et. al. Comput. Phys. Commun 190, 33-50, 2015
        # https://pure.mpg.de/rest/items/item_2085135_9/component/file_2156800/content
        displacement = torch.zeros(
            (num_batch, 3, 3),
            dtype=pos.dtype,
            device=pos.device,
        )
        displacement.requires_grad_(True)
        data["_displacement"] = displacement
        # in the above paper, the infinitesimal distortion is *symmetric*
        # so we symmetrize the displacement before applying it to
        # the positions/cell
        # This is not strictly necessary (reasoning thanks to Mario):
        # the displacement's asymmetric 1o term corresponds to an
        # infinitesimal rotation, which should not affect the final
        # output (invariance).
        # That said, due to numerical error, this will never be
        # exactly true. So, we symmetrize the deformation to
        # take advantage of this understanding and not rely on
        # the invariance here:
        symmetric_displacement = 0.5 * (displacement + displacement.transpose(-1, -2))
        did_pos_req_grad: bool = pos.requires_grad
        pos.requires_grad_(True)
        # bmm is natom in batch
        data[AtomicDataDict.POSITIONS_KEY] = pos + torch.bmm(
            pos.unsqueeze(-2), symmetric_displacement[batch]
        ).squeeze(-2)
        # we only displace the cell if we have one:
        if has_cell:
            # bmm is num_batch in batch
            # here we apply the distortion to the cell as well
            # this is critical also for the correctness
            # if we didn't symmetrize the distortion, since without this
            # there would then be an infinitesimal rotation of the positions
            # but not cell, and it thus wouldn't be global and have
            # no effect due to equivariance/invariance.
            data[AtomicDataDict.CELL_KEY] = cell + torch.bmm(
                cell, symmetric_displacement
            )

        # Call model and get gradients
        data = self.func(data)

        grads = torch.autograd.grad(
            [data[AtomicDataDict.TOTAL_ENERGY_KEY].sum()],
            [pos, data["_displacement"]],
            create_graph=self.training,  # needed to allow gradients of this output during training
        )

        # Put negative sign on forces
        forces = grads[0]
        if forces is None:
            # condition needed to unwrap optional for torchscript
            assert False, "failed to compute forces autograd"
        forces = torch.neg(forces)
        data[AtomicDataDict.FORCE_KEY] = forces

        # Store virial
        virial = grads[1]
        if virial is None:
            # condition needed to unwrap optional for torchscript
            assert False, "failed to compute virial autograd"

        # we only compute the stress (1/V * virial) if we have a cell whose volume we can compute
        if has_cell:
            # ^ can only scale by cell volume if we have one...:
            # Rescale stress tensor
            # See https://github.com/atomistic-machine-learning/schnetpack/blob/master/src/schnetpack/atomistic/output_modules.py#L180
            # First dim is batch, second is vec, third is xyz
            volume = torch.einsum(
                "zi,zi->z",
                cell[:, 0, :],
                torch.cross(cell[:, 1, :], cell[:, 2, :], dim=1),
            ).unsqueeze(-1)
            stress = virial / volume.view(-1, 1, 1)
            data[AtomicDataDict.CELL_KEY] = orig_cell
        else:
            stress = self._empty  # torchscript

        data[AtomicDataDict.STRESS_KEY] = stress

        # see discussion in https://github.com/libAtoms/QUIP/issues/227 about sign convention
        # they say the standard convention is virial = -stress x volume
        # looking above this means that we need to pick up another negative sign for the virial
        # to fit this equation with the stress computed above
        virial = torch.neg(virial)
        data[AtomicDataDict.VIRIAL_KEY] = virial

        # Remove helper
        del data["_displacement"]
        if not did_pos_req_grad:
            # don't give later modules one that does
            pos.requires_grad_(False)

        return data



# -*- coding: utf-8 -*-

# @Time  : 2022/9/27 15:12
# @Author : boliqq07
# @Software: PyCharm
# @License: MIT License

import torch
import torch.nn.functional as F
from math import pi as PI
from torch.nn import GRU, Linear, ReLU, Sequential
from torch_geometric.nn.aggr import Set2Set
from torch_geometric.nn.conv import MessagePassing
from torch_geometric.nn.dense.linear import Linear


class ShiftedSoftplus(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.shift = torch.log(torch.tensor(2.0)).item()

    def forward(self, x):
        return F.softplus(x) - self.shift


class CFConv(MessagePassing):
    def __init__(self, in_channels, out_channels, nn, cutoff):
        super().__init__(aggr='mean')
        self.lin1 = Linear(in_channels, in_channels, bias=False)
        self.lin2 = Linear(in_channels, out_channels)
        self.nn = nn
        self.act = ShiftedSoftplus()
        self.cutoff = cutoff
        self.lin3 = Linear(out_channels, out_channels, bias=False)
        self.reset_parameters()

    def reset_parameters(self):
        torch.nn.init.xavier_uniform_(self.lin1.weight)
        torch.nn.init.xavier_uniform_(self.lin2.weight)
        torch.nn.init.xavier_uniform_(self.lin3.weight)
        self.lin2.bias.data.fill_(0)

    def forward(self, x, edge_index, edge_attr, edge_weight):
        C = (torch.cos(edge_weight * PI / self.cutoff) + 1.0) * 0.5
        W = self.nn(edge_attr) * C.view(-1, 1)

        x = self.lin1(x)
        x = self.propagate(edge_index, x=x, W=W)
        x = self.lin2(x)
        x = self.act(x)
        x = self.lin3(x)

        return x

    def message(self, x_j, W):
        return x_j * W


class GaussianSmearing(torch.nn.Module):
    def __init__(self, start=0.0, stop=5.0, num_gaussians=50):
        super().__init__()
        offset = torch.linspace(start, stop, num_gaussians)
        self.coeff = -0.5 / (offset[1] - offset[0]).item() ** 2
        self.register_buffer('offset', offset)

    def forward(self, dist):
        dist = dist.view(-1, 1) - self.offset.view(1, -1)
        return torch.exp(self.coeff * torch.pow(dist, 2))


class CGGRU(torch.nn.Module):
    def __init__(self, nfeat_node=19, nc_edge_hidden=50, dim=128, cutoff=5.0, **kwargs):
        super().__init__()
        """nfeat_edge=None, nfeat_state=0"""

        self.lin0 = torch.nn.Linear(nfeat_node, dim)
        self.distance_expansion = GaussianSmearing(0.0, cutoff, nc_edge_hidden)

        nn = Sequential(Linear(nc_edge_hidden, 128), ReLU(), Linear(128, dim))
        self.conv = CFConv(dim, dim, nn, cutoff)
        self.gru = GRU(dim, dim)

        self.set2set = Set2Set(dim, processing_steps=3)
        self.lin1 = torch.nn.Linear(2 * dim, dim)
        self.lin2 = torch.nn.Linear(dim, 1)

    def forward(self, data):
        data.edge_attr = self.distance_expansion(data.edge_weight)

        out = F.relu(self.lin0(data.x))
        h = out.unsqueeze(0)

        for i in range(3):
            m = F.relu(self.conv(out, data.edge_index, data.edge_attr, data.edge_weight))
            out, h = self.gru(m.unsqueeze(0), h)
            out = out.squeeze(0)

        out = self.set2set(out, data.batch)
        out = F.relu(self.lin1(out))
        out = self.lin2(out)
        return out.view(-1)



# -*- coding: utf-8 -*-

# @Time  : 2022/9/10 16:26
# @Author : boliqq07
# @Software: PyCharm
# @License: MIT License
from torch import nn
from torch.nn import Module, ModuleList, BatchNorm1d

from hpc2ml.nn.templatemodel import get_active_layer


class ResBlockSameSize(Module):
    """

    ResN(nn.Linear,n_res=2,in_features=64,out_features=64)

    """

    def __init__(self, m, n_res=1, batch_norm=True, active="ReLU", force_x_match=False, **kwargs):
        super().__init__()

        res_layer = ModuleList()

        for _ in range(n_res):
            res_layer.append(m(**kwargs))
            if batch_norm is True:
                bm = BatchNorm1d(num_features=list(kwargs.values())[1])  # num_features = out_features
                res_layer.append(bm)
            res_layer.append(get_active_layer(active)())

        self.first_layer = res_layer[0]
        self.middle_layer = res_layer[1:-1]  # del first and last active layer
        self.force_x_match = force_x_match
        if force_x_match:
            self.x_match_layer = nn.Linear(list(kwargs.values())[0], list(kwargs.values())[1])
        self.last_active_layer = res_layer[-1]

        self.n_res = n_res
        self.block_size = 3 if batch_norm else 2

    def reset_parameters(self):
        self.first_layer.reset_parameters()
        self.middle_layer.reset_parameters()

    def forward(self, h, data=None):

        out = self.first_layer(h, data=data)

        for n, ci in enumerate(self.middle_layer):
            if (n + 1) % self.block_size == 0:
                out = ci(out, data=data)  # m.
            else:
                out = ci(out)  # active or relu.

        if self.force_x_match:
            h = self.x_match_layer(h)

        out = self.last_active_layer(out + h)

        return out


class ResBlockDiffSize(Module):
    """

    ResBlockDiffSize(nn.Linear,layer_size_seq=(64,128,256))

    """

    def __init__(self, m, layer_size_seq=(64, 128, 256), batch_norm=True, active="ReLU",
                 force_x_match=False, **kwargs):
        super().__init__()
        assert len(layer_size_seq) >= 2

        res_layer = ModuleList()

        for i in range(len(layer_size_seq) - 1):
            res_layer.append(m(layer_size_seq[i], layer_size_seq[i + 1], **kwargs))  # in,out, others.
            if batch_norm is True:
                bm = BatchNorm1d(num_features=layer_size_seq[i + 1])  # num_features = out_features
                res_layer.append(bm)
            res_layer.append(get_active_layer(active)())

        self.first_layer = res_layer[0]
        self.middle_layer = res_layer[1:-1]  # del first and last active layer
        self.force_x_match = force_x_match
        if force_x_match:
            self.x_match_layer = nn.Linear(layer_size_seq[0], layer_size_seq[-1])
        self.last_active_layer = res_layer[-1]

        self.n_res = len(layer_size_seq) - 1
        self.block_size = 3 if batch_norm else 2

    def reset_parameters(self):
        self.first_layer.reset_parameters()
        self.middle_layer.reset_parameters()

    def forward(self, h, data=None):

        out = self.first_layer(h, data=data)

        for n, ci in enumerate(self.middle_layer):
            if (n + 1) % self.block_size == 0:
                out = ci(out, data=data)  # m.
            else:
                out = ci(out)  # active or relu.

        if self.force_x_match:
            h = self.x_match_layer(h)

        out = self.last_active_layer(out + h)

        return out



"""
Utilities that involve file searching and operations (i.e. save/load)
"""
import contextlib
import contextvars
import logging
import os
import shutil
import sys
import tempfile
from pathlib import Path
from typing import Union, List, Tuple, Optional, Callable

import yaml

# accumulate writes to group for renaming
_MOVE_SET = contextvars.ContextVar("_move_set", default=None)


def _delete_files_if_exist(paths):
    # clean up
    # better for python 3.8 >
    if sys.version_info[1] >= 8:
        for f in paths:
            f.unlink(missing_ok=True)
    else:
        # race condition?
        for f in paths:
            if f.exists():
                f.unlink()


def _process_moves(moves: List[Tuple[bool, Path, Path]]):
    """blocking to copy (possibly across filesystems) to temp name; then atomic rename to final name"""
    try:
        for _, from_name, to_name in moves:
            # blocking copy to temp file in same filesystem
            tmp_path = to_name.parent / (f".tmp-{to_name.name}~")
            shutil.move(from_name, tmp_path)
            # then atomic rename to overwrite
            tmp_path.rename(to_name)
    finally:
        _delete_files_if_exist([m[1] for m in moves])


# allow user to enable/disable depending on their filesystem
_ASYNC_ENABLED = os.environ.get("NEQUIP_ASYNC_IO", "false").lower()
assert _ASYNC_ENABLED in ("true", "false")
_ASYNC_ENABLED = _ASYNC_ENABLED == "true"

if _ASYNC_ENABLED:
    import threading
    from queue import Queue

    _MOVE_QUEUE = Queue()
    _MOVE_THREAD = None


    # Because we use a queue, later writes will always (correctly)
    # overwrite earlier writes
    def _moving_thread(queue):
        while True:
            moves = queue.get()
            _process_moves(moves)
            # logging is thread safe: https://stackoverflow.com/questions/2973900/is-pythons-logging-module-thread-safe
            logging.debug(f"Finished writing {', '.join(m[2].name for m in moves)}")
            queue.task_done()


    def _submit_move(from_name, to_name, blocking: bool):
        global _MOVE_QUEUE
        global _MOVE_THREAD
        global _MOVE_SET

        # launch thread if its not running
        if _MOVE_THREAD is None:
            _MOVE_THREAD = threading.Thread(
                target=_moving_thread, args=(_MOVE_QUEUE,), daemon=True
            )
            _MOVE_THREAD.start()

        # check on health of copier thread
        if not _MOVE_THREAD.is_alive():
            _MOVE_THREAD.join()  # will raise exception
            raise RuntimeError("Writer thread failed.")

        # submit this move
        obj = (blocking, from_name, to_name)
        if _MOVE_SET.get() is None:
            # no current group
            _MOVE_QUEUE.put([obj])
            # if it should be blocking, wait for it to be processed
            if blocking:
                _MOVE_QUEUE.join()
        else:
            # add and let the group submit and block (or not)
            _MOVE_SET.get().append(obj)


    @contextlib.contextmanager
    def atomic_write_group():
        global _MOVE_SET
        if _MOVE_SET.get() is not None:
            # nesting is a no-op
            # submit along with outermost context manager
            yield
            return
        token = _MOVE_SET.set(list())
        # run the saves
        yield
        _MOVE_QUEUE.put(_MOVE_SET.get())  # send it off
        # if anyone is blocking, block the whole group:
        if any(m[0] for m in _MOVE_SET.get()):
            # someone is blocking
            _MOVE_QUEUE.join()
        # exit context
        _MOVE_SET.reset(token)


    def finish_all_writes():
        global _MOVE_QUEUE
        _MOVE_QUEUE.join()
        # ^ wait for all remaining moves to be processed

else:

    def _submit_move(from_name, to_name, blocking: bool):
        global _MOVE_SET
        obj = (blocking, from_name, to_name)
        if _MOVE_SET.get() is None:
            # no current group just do it
            _process_moves([obj])
        else:
            # add and let the group do it
            _MOVE_SET.get().append(obj)


    @contextlib.contextmanager
    def atomic_write_group():
        global _MOVE_SET
        if _MOVE_SET.get() is not None:
            # don't nest them
            yield
            return
        token = _MOVE_SET.set(list())
        yield
        _process_moves(_MOVE_SET.get())  # do it
        _MOVE_SET.reset(token)


    def finish_all_writes():
        pass  # nothing to do since all writes blocked


@contextlib.contextmanager
def atomic_write(
        filename: Union[Path, str, List[Union[Path, str]]],
        blocking: bool = True,
        binary: bool = False,
):
    aslist: bool = True
    if not isinstance(filename, list):
        aslist = False
        filename = [filename]
    filename = [Path(f) for f in filename]

    with contextlib.ExitStack() as stack:
        files = [
            stack.enter_context(
                tempfile.NamedTemporaryFile(
                    mode="w" + ("b" if binary else ""), delete=False
                )
            )
            for _ in filename
        ]
        try:
            if not aslist:
                yield files[0]
            else:
                yield files
        except:  # noqa
            # ^ noqa cause we want to delete them no matter what if there was a failure
            # only remove them if there was an error
            _delete_files_if_exist([Path(f.name) for f in files])
            raise

        for tp, fname in zip(files, filename):
            _submit_move(Path(tp.name), Path(fname), blocking=blocking)


def save_file(
        item,
        supported_formats: dict,
        filename: str,
        enforced_format: str = None,
        blocking: bool = True,
):
    """
    Save file. It can take yaml, json, pickle, json, npz and torch save
    """

    # check whether folder exist
    path = os.path.dirname(os.path.realpath(filename))
    if not os.path.isdir(path):
        logging.debug(f"save_file make dirs {path}")
        os.makedirs(path, exist_ok=True)

    format, filename = adjust_format_name(
        supported_formats=supported_formats,
        filename=filename,
        enforced_format=enforced_format,
    )

    with atomic_write(
            filename,
            blocking=blocking,
            binary={
                "json": False,
                "yaml": False,
                "pickle": True,
                "torch": True,
                "npz": True,
            }[format],
    ) as write_to:
        if format == "json":
            import json

            json.dump(item, write_to)
        elif format == "yaml":
            import yaml

            yaml.dump(item, write_to)
        elif format == "torch":
            import torch

            torch.save(item, write_to)
        elif format == "pickle":
            import pickle

            pickle.dump(item, write_to)
        elif format == "npz":
            import numpy as np

            np.savez(write_to, item)
        else:
            raise NotImplementedError(
                f"Output format {format} not supported:"
                f" try from {supported_formats.keys()}"
            )

    return filename


def load_file(supported_formats: dict, filename: str, enforced_format: str = None):
    """
    Load file. Current support form
    """
    if enforced_format is None:
        format = match_suffix(supported_formats=supported_formats, filename=filename)
    else:
        format = enforced_format

    if not os.path.isfile(filename):
        abs_path = str(Path(filename).resolve())
        raise OSError(f"file {filename} at {abs_path} is not found")

    if format == "json":
        import json

        with open(filename) as fin:
            return json.load(fin)
    elif format == "yaml":
        import yaml

        with open(filename) as fin:
            return yaml.load(fin, Loader=yaml.Loader)
    elif format == "torch":
        import torch

        return torch.load(filename)
    elif format == "pickle":
        import pickle

        with open(filename, "rb") as fin:
            return pickle.load(fin)
    elif format == "npz":
        import numpy as np

        return np.load(filename, allow_pickle=True)
    else:
        raise NotImplementedError(
            f"Input format not supported:" f" try from {supported_formats.keys()}"
        )


def load_callable(obj: Union[str, Callable], prefix: Optional[str] = None) -> Callable:
    """Load a callable from a name, or pass through a callable."""
    if callable(obj):
        pass
    elif isinstance(obj, str):
        if "." not in obj:
            # It's an unqualified name
            if prefix is not None:
                obj = prefix + "." + obj
            else:
                # You can't have an unqualified name without a prefix
                raise ValueError(f"Cannot load unqualified name {obj}.")
        obj = yaml.load(f"!!python/name:{obj}", Loader=yaml.Loader)
    else:
        raise TypeError
    assert callable(obj), f"{obj} isn't callable"
    return obj


def adjust_format_name(
        supported_formats: dict, filename: str, enforced_format: str = None
):
    """
    Recognize whether proper suffix is added to the filename.
    If not, add it and return the formatted file name

    Args:

        supported_formats (dict): list of supported formats and corresponding suffix
        filename (str): initial filename
        enforced_format (str): default format

    Returns:

        newformat (str): the chosen format
        newname (str): the adjusted filename

    """
    if enforced_format is None:
        newformat = match_suffix(supported_formats=supported_formats, filename=filename)
    else:
        newformat = enforced_format

    newname = f"{filename}"

    add_suffix = True
    suffix = supported_formats[newformat]

    if not isinstance(suffix, (set, list, tuple)):
        suffix = [suffix]

    if len(suffix) > 0:
        for suf in suffix:
            if filename.endswith(f".{suf}"):
                add_suffix = False

        if add_suffix:
            suffix = suffix[0]
            newname += f".{suffix}"

    return newformat, newname


def match_suffix(supported_formats: str, filename: str):
    """
    Recognize format based on suffix

    Args:

        supported_formats (dict): list of supported formats and corresponding suffix
        filename (str): initial filename

    Returns:

        format (str): the recognized format

    """
    for form, suffs in supported_formats.items():
        if isinstance(suffs, (set, list, tuple)):
            for suff in suffs:
                if filename.lower().endswith(f".{suff}"):
                    return form
        else:
            if filename.lower().endswith(f".{suffs}"):
                return form

    return list(supported_formats.keys())[0]



from torch import nn

from hpc2ml.nn.activations import Act


class Forces(nn.Module):
    """Forces net"""

    def __init__(self, output_dim, decoder_type="mlp", decoder_activation_str="ssp", ):
        super(Forces, self).__init__()
        self.decoder_type = decoder_type
        self.act = Act(decoder_activation_str)
        self.output_dim = output_dim

        if self.decoder_type == "linear":
            self.decoder = nn.Sequential(nn.Linear(self.output_dim, 3))
        elif self.decoder_type == "mlp":
            self.decoder = nn.Sequential(
                nn.Linear(self.output_dim, self.output_dim),
                self.act,
                nn.Linear(self.output_dim, int(self.output_dim / 1)),
                self.act,
                nn.Linear(int(self.output_dim / 1), 3),
            )
        else:
            raise ValueError(f"Undefined forces decoder: {self.decoder_type}")

        self.reset_parameters()

    def reset_parameters(self):
        for m in self.decoder:
            if isinstance(m, nn.Linear):
                nn.init.xavier_uniform_(m.weight)
                m.bias.data.fill_(0)

        self.decoder[-1].weight.data = self.decoder[-1].weight.data / 10

    def forward(self, x):
        x = self.decoder(x)
        return x



# -*- coding: utf-8 -*-

# @Time  : 2023/2/2 11:22
# @Author : boliqq07
# @Software: PyCharm
# @License: MIT License
"""This part is from ase.optimize and add more function."""

import warnings

import numpy as np
from ase.optimize.gpmin.kernel import SquaredExponential
from ase.optimize.gpmin.prior import ConstantPrior
from ase.optimize.optimize import Optimizer
from ase.parallel import world
from ase.utils import write_json
from scipy import stats
from scipy.optimize import minimize

from hpc2ml.optimize.gp import GaussianProcess


class EISupport:
    # More
    def __init__(self, atoms=None, converged_std=None):
        self.atoms = atoms
        self.converged_std = converged_std
        self.force_consistent = False

    def _converged_force(self, forces=None):
        """Did the optimization converge?"""
        if forces is None:
            forces = self.atoms.get_forces()
        return (forces ** 2).sum(axis=1).max() < self.fmax ** 2

    def _converged_energy_range(self, forces=None):
        """Did the optimization converge?"""
        _ = forces
        return len(self.es) < 3 or np.std(self.es) > 0.01

    def converged(self, forces=None):
        if self.converged_std == "forces":
            return self._converged_force(forces=forces)
        elif self.converged_std == "energy_range":
            return self._converged_energy_range(forces=forces)
        elif self.converged_std == "default":
            return self._converged(forces=forces)
        else:
            return self._converged_force(forces=forces)

    def _get_pos_r_e_f(self, atoms_list):

        e = []
        f = []
        r = []

        calc = atoms_list[0].calc

        if hasattr(calc, "calculate_batch"):
            res = calc.calculate_batch(atoms_list)
            r = np.array([i.positions.ravel() for i in atoms_list])
            e = res["energy"]
            f = res["forces"]
            index = np.isfinite(e)
            r = r[index]
            e = e[index]
            f = f[index]
        else:
            for atom_temp in atoms_list:

                ei = atom_temp.get_potential_energy(force_consistent=self.force_consistent)
                fi = atom_temp.get_forces().ravel()
                ri = atom_temp.positions.ravel()

                if np.isfinite(ei):
                    e.append(ei)
                    f.append(fi.reshape(1, -1))
                    r.append(ri.reshape(1, -1))

        if len(e) > 0:
            e = np.array(e).reshape(-1, 1)
            f = np.concatenate(f, axis=0) if not isinstance(f, np.ndarray) else f
            r = np.concatenate(r, axis=0) if not isinstance(r, np.ndarray) else r
        else:
            r = self.atoms.positions.reshape(1, -1)
            f = self.atoms.get_forces().reshape(1, -1)
            e = self.atoms.get_potential_energy(force_consistent=self.force_consistent)
            e = np.array(e).reshape(-1, 1)
        return r, e, f

    @staticmethod
    def _get_pos_r(atom_list):
        """and gather pos matrix from a batch of atoms """
        r = [i.positions.reshape(1, -1) for i in atom_list]
        r = np.concatenate(r, axis=0)
        return r

    def resample_atom(self, n, random_factor=0.001, axis_factor=(1, 1, 1)):
        """resample atoms."""
        # random_state = 0
        np.random.set_state(np.random.get_state())
        atoms_list = []
        for i in range(n):
            atom_temp = self.atoms.copy()
            atom_temp.calc = self.atoms.calc
            st_frac_coords = atom_temp.get_scaled_positions()
            axis_factor = np.array(list(axis_factor)).reshape(1, -1)
            st_frac_coords = np.array(st_frac_coords) + (
                    np.random.random(st_frac_coords.shape) - 0.5) * random_factor * axis_factor
            atom_temp.set_scaled_positions(st_frac_coords)
            atoms_list.append(atom_temp)

        return atoms_list

    def resample_r(self, n, random_factor=0.001, axis_factor=(1, 1, 1)):
        """resample atoms pos."""
        atom_list = self.resample_atom(n, random_factor=random_factor,
                                       axis_factor=axis_factor)
        r = self._get_pos_r(atom_list)
        return r

    def resample_r_e_f(self, n, random_factor=0.001, axis_factor=(1, 1, 1)):
        """resample atoms pos,energy,forces."""
        atom_list = self.resample_atom(n, random_factor=random_factor,
                                       axis_factor=axis_factor)
        r, e, f = self._get_pos_r_e_f(atom_list)
        return r, e, f

    def r_to_e_f(self, r):

        atoms_list = []
        for i in range(r.shape[0]):
            atom_temp = self.atoms.copy()
            atom_temp.calc = self.atoms.calc
            atom_temp.set_positions(r[i].reshape(-1, 3))
            atoms_list.append(atom_temp)

        r, e, f = self._get_pos_r_e_f(atoms_list)

        return r, e, f

    def r_to_e_f_single(self, ri, inplace=False):
        if not inplace:
            atom_temp = self.atoms.copy()
            atom_temp.calc = self.atoms.calc
        else:
            atom_temp = self.atoms

        atom_temp.set_positions(ri.reshape(-1, 3))
        ei = atom_temp.get_potential_energy(force_consistent=self.force_consistent)
        fi = atom_temp.get_forces().ravel()
        ri = ri.ravel()
        return ri, ei, fi

    def predict_mean_std(self, x):
        """This function should re-defined"""
        f, v = self.predict(x, get_variance=True) # predict is just for GP
        # e_mean = f[0]
        # f_mean = f[1:]
        v2 = np.diag(v)
        # e_std = v2[0]
        # f_std = v2[1:]
        e_mean_f_mean_e_std_f_std = np.vstack((f.ravel(), v2.ravel()))
        return e_mean_f_mean_e_std_f_std

    def predict_mean_std_all(self, xs):
        """e_mean_f_mean_e_std_f_std.
        shape (n_sample, 2 [mean,std], n_atom+1 [e_ms,f_ms] )"""
        res = [self.predict_mean_std(x) for x in xs]
        return np.array(res)

    @staticmethod
    def ei(mean_std, sign=1, target=None):

        mean = mean_std[:, 0]
        std = mean_std[:, 1]
        zero_std_index = (std < 0.001) & (std > -0.001)
        no_zero_std_index = ~zero_std_index

        # zero_std_index = std == 0.0
        # no_zero_std_index = std != 0.0
        ei = np.zeros_like(mean)

        if target is not None:

            upper = mean > target
            # lower = mean <= target

            mean[upper] = mean[upper] * -1  # trans value more than target to min problem

            mz = mean[zero_std_index]
            # sz = std[zero_std_index]

            if np.any(zero_std_index):
                ei0 = (mz - target) * 0.5  # 0.5 is stats.norm.cdf(0.0)
                ei[zero_std_index] = ei0

            mnz = mean[no_zero_std_index]
            snz = std[no_zero_std_index]

            if np.any(no_zero_std_index):
                kg = (mnz - target) / snz
                ei1 = mnz * stats.norm.cdf(kg) + snz * stats.norm.pdf(kg)
                ei[no_zero_std_index] = ei1

        else:
            mean = mean * sign  # trans min problem to max problem

            mz = mean[zero_std_index]
            # sz = std[zero_std_index]
            if np.any(zero_std_index):
                ei0 = (mz - np.max(mean)) * 0.5  # 0.5 is stats.norm.cdf(0.0)
                ei[zero_std_index] = ei0

            mnz = mean[no_zero_std_index]
            snz = std[no_zero_std_index]

            if np.any(no_zero_std_index):
                kg = (mnz - np.max(mean)) / snz
                ei1 = mnz * stats.norm.cdf(kg) + snz * stats.norm.pdf(kg)
                ei[no_zero_std_index] = ei1

        return ei

    def ei_balance(self, mean_std, loss_coef=1):

        fm = 0.0
        fn = mean_std.shape[2]

        ef_balance_coef = 1

        coef = loss_coef / (fn - 1) * ef_balance_coef

        ei_all = np.zeros((mean_std.shape[0]), dtype=np.float64)

        for i in range(fn):
            msi = mean_std[:, :, i]
            if i == 0:
                eii = self.ei(msi, sign=-1)  # less
            else:
                eii = self.ei(msi, target=fm)
                # more near to 0
            eii = eii.ravel()
            if np.max(eii) == np.min(eii):
                eii = 0.0
            else:
                eii = (eii - np.min(eii)) / (np.max(eii) - np.min(eii))
            if i > 0:
                eii = eii * coef
            ei_all += eii

        return ei_all.ravel()


class BaseGPMin(EISupport, Optimizer, GaussianProcess):
    def __init__(self, atoms, restart=None, logfile='-', trajectory=None,
                 prior=None, kernel=None, master=None, noise=None, weight=None,
                 scale=None, force_consistent=None, batch_size=None,
                 bounds=None, update_prior_strategy="maximum",
                 update_hyperparams=False, converged_std="forces"):

        # Warn the user if the number of atoms is very large
        if len(atoms) > 100:
            warning = ('Possible Memory Issue. There are more than '
                       '100 atoms in the unit cell. The memory '
                       'of the process will increase with the number '
                       'of steps, potentially causing a memory issue. '
                       'Consider using a different optimizer.')

            warnings.warn(warning)

        # Give it default hyperparameters
        if update_hyperparams:  # Updated GPMin
            if scale is None:
                scale = 0.3
            if noise is None:
                noise = 0.004
            if weight is None:
                weight = 2.
            if bounds is None:
                self.eps = 0.1
            elif bounds is False:
                self.eps = None
            else:
                self.eps = bounds
            if batch_size is None:
                self.nbatch = 1
            else:
                self.nbatch = batch_size
        else:  # GPMin without updates
            if scale is None:
                scale = 0.4
            if noise is None:
                noise = 0.001
            if weight is None:
                weight = 1.
            if bounds is not None:
                warning = ('The parameter bounds is of no use '
                           'if update_hyperparams is False. '
                           'The value provided by the user '
                           'is being ignored.')
                warnings.warn(warning, UserWarning)
            if batch_size is not None:
                warning = ('The parameter batch_size is of no use '
                           'if update_hyperparams is False. '
                           'The value provided by the user '
                           'is being ignored.')
                warnings.warn(warning, UserWarning)

            # Set the variables to something anyways
            self.eps = False
            self.nbatch = None

        self.strategy = update_prior_strategy
        self.update_hp = update_hyperparams
        self.function_calls = 1
        self.force_calls = 0
        self.x_list = []  # Training set features
        self.y_list = []  # Training set targets

        Optimizer.__init__(self, atoms, restart, logfile,
                           trajectory, master, force_consistent)
        if prior is None:
            self.update_prior = True
            prior = ConstantPrior(constant=None)
        else:
            self.update_prior = False

        if kernel is None:
            kernel = SquaredExponential()
        GaussianProcess.__init__(self, prior, kernel)
        self.set_hyperparams(np.array([weight, scale, noise]))
        EISupport.__init__(self, atoms=atoms, converged_std=converged_std)

    def update(self, r, e, f):
        """Update the PES

        Update the training set, the prior and the hyperparameters.
        Finally, train the model
        """
        # update the training set
        self.x_list.append(r)
        f = f.reshape(-1)
        y = np.append(np.array(e).reshape(-1), -f)
        self.y_list.append(y)
        
        if len(self.y_list) > 32:
            self.x_list = self.x_list[-32:]
            self.y_list = self.y_list[-32:]

        # Set/update the constant for the prior
        if self.update_prior:
            if self.strategy == 'average':
                av_e = np.mean(np.array(self.y_list)[:, 0])
                self.prior.set_constant(av_e)
            elif self.strategy == 'maximum':
                max_e = np.max(np.array(self.y_list)[:, 0])
                self.prior.set_constant(max_e)

        # update hyperparams
        if (self.update_hp and self.function_calls % self.nbatch == 0 and
                self.function_calls != 0):
            self.fit_to_batch()

        # build the model
        self.train(np.array(self.x_list), np.array(self.y_list))

    def update_batch(self, rs, es, fs):
        """Update the PES

        Update the training set, the prior and the hyperparameters.
        Finally, train the model
        """
        # update the training set
        [self.x_list.append(ri) for ri in rs]
        y = np.concatenate((np.array(es).reshape(-1, 1), -fs), axis=1)
        [self.y_list.append(yi) for yi in y]
        if len(self.y_list) > 32:
            self.x_list = self.x_list[-32:]
            self.y_list = self.y_list[-32:]

        # Set/update the constant for the prior
        if self.update_prior:
            if self.strategy == 'average':
                av_e = np.mean(np.array(self.y_list)[:, 0])
                self.prior.set_constant(av_e)
            elif self.strategy == 'maximum':
                max_e = np.max(np.array(self.y_list)[:, 0])
                self.prior.set_constant(max_e)

        # update hyperparams
        if (self.update_hp and self.function_calls % self.nbatch == 0 and
                self.function_calls != 0):
            self.fit_to_batch()

        # build the model
        self.train(np.array(self.x_list), np.array(self.y_list))

    def remove(self, l):
        """should after update_batch"""

        self.x_list = self.x_list[:-l]
        self.y_list = self.y_list[:-l]

    def acquisition(self, r):
        e = self.predict(r)
        return e[0], e[1:]

    def fit_to_batch(self):
        """Fit hyperparameters keeping the ratio noise/weight fixed"""
        ratio = self.noise / self.kernel.weight
        self.fit_hyperparameters(np.array(self.x_list),
                                 np.array(self.y_list), eps=self.eps)
        self.noise = ratio * self.kernel.weight

    def relax_model(self, r0):
        result = minimize(self.acquisition, r0, method='L-BFGS-B', jac=True)
        if result.success:
            return result.x
        else:
            self.dump()
            warnings.warn("The minimization of the acquisition function "
                               "has not converged")
            return result.x

    def dump(self):
        """Save the training set"""
        if world.rank == 0 and self.restart is not None:
            with open(self.restart, 'wb') as fd:
                write_json(fd, (self.x_list, self.y_list))

    def read(self):
        self.x_list, self.y_list = self.load()

    def step(self, f=None):
        atoms = self.atoms
        if f is None:
            f = atoms.get_forces()

        fc = self.force_consistent
        r0 = self.atoms.get_positions().reshape(-1)
        e0 = self.atoms.get_potential_energy(force_consistent=fc)
        self.update(r0, e0, f)

        r1 = self.relax_model(r0)
        self.atoms.set_positions(r1.reshape(-1, 3))
        e1 = self.atoms.get_potential_energy(force_consistent=fc)
        f1 = self.atoms.get_forces()
        self.function_calls += 1
        self.force_calls += 1
        count = 0
        while e1 >= e0:
            self.update(r1, e1, f1)
            r1 = self.relax_model(r0)

            self.atoms.set_positions(r1.reshape(-1, 3))
            e1 = self.atoms.get_potential_energy(force_consistent=fc)
            f1 = self.atoms.get_forces()
            self.function_calls += 1
            self.force_calls += 1
            if self.converged(f1):
                break

            count += 1
            if count == 10:
                warnings.warn("A descent model could not be built")
                break
        self.dump()



# -*- coding: utf-8 -*-
import os
from typing import List

import numpy as np
import path
import torch
from ase import Atoms
from ase.calculators.calculator import Calculator, PropertyNotImplementedError

from hpc2ml.data.batchdata import MtBatchData
from hpc2ml.data.structuretodata import StructureToData
from hpc2ml.nn.flow_geo import simple_predict
from hpc2ml.nn.metrics import mlm


class GNNEICalculater(Calculator):
    implemented_properties: List[str] = ["energy", "forces"]

    _deprecated = object()

    'Properties calculator can handle (energy, forces)'

    def __init__(self, model, resume_file, atoms=None, directory='.', convert=StructureToData(),
                 properties=None,
                 device="cpu",
                 **kwargs):
        """Basic calculator implementation.

        restart: str
            Prefix for restart file.  May contain a directory. Default
            is None: don't restart.

        directory: str or PurePath
            Working directory in which to read and write files and
            perform calculations.
        label: str
            Name used for all files.  Not supported by all calculators.
            May contain a directory, but please use the directory parameter
            for that instead.
        atoms: Atoms object
            Optional Atoms object to which the calculator will be
            attached.  When restarting, atoms will get its positions and
            unit-cell updated from file.
        """
        super(GNNEICalculater, self).__init__(atoms, directory=directory, **kwargs)

        self.model = model
        self.convert = convert
        self.convert.tq = False
        self.resume_file = resume_file
        self.device = torch.device(device)
        if properties is not None:
            self.implemented_properties = properties

        from hpc2ml.nn.flow_geo import load_check_point

        resume_file = path.Path(self.directory) / self.resume_file

        optimizer = torch.optim.Adam(self.model.parameters(), lr=0.001)
        model, optimizer, start_epoch, best_error, note = load_check_point(self.model,
                                                                           resume_file=resume_file,
                                                                           optimizer=optimizer,
                                                                           device=self.device)

        self.model = model
        self.optimizer = optimizer
        self.note = note
        self.start_epoch = start_epoch
        self.best_error = best_error

    def train(self, atoms, batch_data=None, **kwargs):
        """train"""
        print("Re-train model...")

        from torch_geometric.loader import DataLoader
        from hpc2ml.nn.flow_geo import LearningFlow

        if batch_data is None:
            dataset = self.get_bath_data(atoms, **kwargs)
        else:
            dataset = batch_data

        test_loader = DataLoader(dataset, batch_size=32, shuffle=False)

        train_loader = DataLoader(dataset, batch_size=32, shuffle=False)
        # # LearningFlow
        print_what = ""
        self.lf = LearningFlow(self.model, train_loader, test_loader=test_loader,
                               loss_method=mlm, optimizer=self.optimizer,
                               device=self.device, note=self.note, multi_loss=True,
                               target_name=tuple(self.implemented_properties),
                               checkpoint=True, store_filename=self.resume_file,
                               loss_threshold=0.001, print_what=print_what,
                               )

        self.lf.start_epoch = self.start_epoch
        self.lf.best_error = self.best_error
        self.lf.run(epoch=5)
        print("Done.")

    def get_bath_data(self, atoms=None, **kwargs) -> "MtBatchData":
        """Get scaled data."""
        batch = MtBatchData.from_atoms(atoms, **kwargs, convert=self.convert)
        batch.scale(dct=self.note)
        return batch

    def get_res_from_bath_data(self, batch: MtBatchData, msg_dict: dict) -> "MtBatchData":
        """Get unscaled data."""

        for k, v in msg_dict.items():
            ### energy and stress is atom
            slice_type = "sample"
            if k == "forces":
                slice_type = "atom"

            batch.add_prop(k, v, slice_type=slice_type)

        batch.unscale(dct=self.note)
        return batch

    def get_property(self, name, atoms=None, allow_calculation=True):
        if name not in self.implemented_properties:
            raise PropertyNotImplementedError('{} property not implemented'
                                              .format(name))

        if atoms is None:
            atoms = self.atoms
        else:
            system_changes = self.check_state(atoms)
            if system_changes:
                self.reset()
        if name not in self.results:
            if not allow_calculation:
                return None
            self.calculate(atoms)

        if name not in self.results:
            # For some reason the calculator was not able to do what we want,
            # and that is OK.
            raise PropertyNotImplementedError('{} not present in this '
                                              'calculation'.format(name))

        result = self.results[name]
        if isinstance(result, np.ndarray):
            result = result.copy()
        return result

    def calculate(self, atoms=None, batch_data=None):
        assert isinstance(atoms, Atoms)
        properties = tuple(self.implemented_properties)

        if atoms is not None:
            self.atoms = atoms.copy()

        if not os.path.isdir(self._directory):
            os.makedirs(self._directory)

        from torch_geometric.loader import DataLoader

        if batch_data is None:
            batch = self.get_bath_data(atoms)
        else:
            batch = batch_data

        batch.to(self.device)

        predict_loader = DataLoader(batch, batch_size=32, shuffle=False)  # must not shuffle

        res = simple_predict(self.model, predict_loader, return_y_true=False, device=self.device,
                             process_out=None, process_label=None, target_name=properties,
                             multi_loss=True)

        res_dict = {k: v for k, v in zip(properties, res)}

        res_batch = self.get_res_from_bath_data(batch, res_dict)

        for k in properties:
            v = getattr(res_batch, k)

            if isinstance(v, torch.Tensor):
                v = v.detach().cpu().numpy()

            v = v.reshape(1, -1)

            if hasattr(v, "shape") and sum(v.shape) <= 2:
                v = v[0]

            self.results.update({k: v})

        if "stress" not in properties:
            self.results["stress"] = np.zeros((1, 6))
        return self.results

    def calculate_batch(self, atoms, batch_data=None):
        properties = tuple(self.implemented_properties)


        if isinstance(atoms, Atoms):
            atoms = [atoms, ]
        else:
            atoms = atoms

        if not os.path.isdir(self._directory):
            os.makedirs(self._directory)

        from torch_geometric.loader import DataLoader

        if batch_data is None:
            batch = self.get_bath_data(atoms)
        else:
            batch = batch_data

        l = len(atoms)

        batch.to(self.device)

        predict_loader = DataLoader(batch, batch_size=32, shuffle=False)

        res_batch = simple_predict(self.model, predict_loader, return_y_true=False,
                                   device=self.device, process_out=None,
                                   process_label=None, target_name=properties,
                                   multi_loss=True)

        res_dict = {k: v for k, v in zip(properties, res_batch)}

        res_batch = self.get_res_from_bath_data(batch, res_dict)

        res_list = []

        for i in range(l):
            results = {}
            res_batch_i = res_batch[i]
            for k in properties:
                v = getattr(res_batch_i, k)

                if isinstance(v, torch.Tensor):
                    v = v.detach().cpu().numpy()

                v = v.reshape(1, -1)

                results.update({k: v})

            if "stress" not in properties:
                results["stress"] = np.zeros((1, 6))
            res_list.append(results)
        return res_list



import numpy as np
from pymatgen.io.vasp import Vasprun


def sparse_vasprun(path: str, forces=True):
    """Sparse vasprun.xml and load finall data."""
    vr = Vasprun(path)
    steps = vr.ionic_steps
    sym = [i.split(" ")[1].split("_")[0] for i in vr.potcar_symbols]
    potcar_label = tuple([vr.potcar_symbols[sym.index(i)] for i in vr.atomic_symbols])

    if forces:
        stepi = steps[-1]
        dct = {f"{path}": {"structure": stepi["structure"], "forces": np.array(stepi["forces"]),
                           "stress": np.array(stepi["stress"]),
                           "energy": stepi["e_wo_entrp"], "potcar_symbol": potcar_label, "incar": vr.incar,
                           "kpoints_kpts": vr.kpoints.kpts, "kpoints_style": vr.kpoints.style.name,
                           "efermi": vr.efermi}}
    else:
        dct = {f"{path}": {"structure": vr.final_structure, "energy": vr.final_energy, "potcar_symbol": potcar_label,
                           "incar": vr.incar, "kpoints_kpts": vr.kpoints.kpts, "kpoints_style": vr.kpoints.style.name,
                           "efermi": vr.efermi}}

    return dct


def sparse_vasprun_traj(path: str, space=20):
    """Sparse vasprun.xml and load all ion step, where space is the sample interval."""
    vr = Vasprun(path)
    steps = vr.ionic_steps
    sym = [i.split(" ")[1].split("_")[0] for i in vr.potcar_symbols]
    potcar_label = tuple([vr.potcar_symbols[sym.index(i)] for i in vr.atomic_symbols])

    dct = {}
    for i, stepi in enumerate(steps):
        if i % space == 0 and i != len(steps) - 1:
            dcti = {"structure": stepi["structure"], "forces": np.array(stepi["forces"]),
                    "stress": np.array(stepi["stress"]),
                    "energy": stepi["e_wo_entrp"], "potcar_symbol": potcar_label, "incar": vr.incar,
                    "kpoints_kpts": vr.kpoints.kpts, "kpoints_style": vr.kpoints.style.name, "efermi": vr.efermi}
            dct.update({f"{path}_step{i}": dcti})

    stepi = steps[-1]
    dctn1 = {f"{path}": {"structure": stepi["structure"], "forces": np.array(stepi["forces"]),
                         "stress": np.array(stepi["stress"]),
                         "energy": stepi["e_wo_entrp"], "potcar_symbol": potcar_label, "incar": vr.incar,
                         "kpoints_kpts": vr.kpoints.kpts, "kpoints_style": vr.kpoints.style.name, "efermi": vr.efermi}}
    dct.update(dctn1)
    return dct


if __name__ == "__main__":
    # path = r"C:\Users\98679\PycharmProjects\hpc2ml\test\test_data\pure_opt\vasprun.xml"
    path = r"../../../../hpc2ml/test/test_data/pure_opt/vasprun.xml"
    path = r"../../../../hpc2ml/test/test_data/OH_add_static/vasprun.xml"
    res = sparse_vasprun_traj(path)
    res2 = sparse_vasprun(path, forces=False)



# -*- coding: utf-8 -*-

# @Time     : 2021/8/29 18:55
# @Software : PyCharm
# @Author   : xxx

"""
This file is used to extension the compare AtomsRow(s).

Example

>>> db = connect("organometal.db")
>>> row1,row2 = db[1],db[2]
>>> atoms_row_matcher(row1,row2)
1.0 # the more close to 1, the more similar.

More compare:
    pymatgen/analysis/structure_matcher.py
    ase/utils/structure_comparator.py
"""

from collections import Iterable
from typing import Dict

import numpy as np
from ase.db.row import AtomsRow


def _equal(ar1: AtomsRow, ar2: AtomsRow, tol=2, resort=True):
    """Just sort site."""
    poa = np.round(ar1.positions, decimals=tol)
    pob = np.round(ar2.positions, decimals=tol)

    if resort:
        poa = np.sort(poa, axis=0)
        poa_index = np.argsort(poa, axis=0)
        pob = np.sort(pob, axis=0)
        pob_index = np.argsort(pob, axis=0)

        return (ar1.natoms == ar2.natoms and
                (poa == pob).all() and (ar1.numbers[poa_index] == ar2.numbers[pob_index]).all())
    else:
        return (ar1.natoms == ar2.natoms and
                (poa == pob).all() and (ar1.numbers == ar2.numbers).all())


def atoms_row_matcher_log(atomsrow1: AtomsRow, atomsrow2: AtomsRow, resort=False) -> Dict:
    """Check the log AtomsRow similarity"""
    result = {"atoms": _equal(atomsrow1, atomsrow2, tol=2, resort=resort)}
    for i in atomsrow1.key_value_pairs.keys() & atomsrow2.key_value_pairs.keys():
        result.update({i: atomsrow1.key_value_pairs[i] == atomsrow2.key_value_pairs[i]})
    for i in atomsrow1.key_value_pairs.keys() - atomsrow2.key_value_pairs.keys():
        result.update({i: False})
    for i in atomsrow2.key_value_pairs.keys() - atomsrow1.key_value_pairs.keys():
        result.update({i: False})
    for i in ['initial_magmoms', 'initial_charges', 'momenta', 'pbc']:
        h1, h2 = hasattr(atomsrow1, i), hasattr(atomsrow2, i)
        if h1 and h2:
            result.update({i: getattr(atomsrow1, i) == getattr(atomsrow2, i)})
        elif not h1 and not h2:
            pass
        else:
            result.update({i: False})

    for i in result.keys():
        if isinstance(result[i], np.ndarray):
            result[i] = np.all(result[i])

    return result


def atoms_row_matcher(atomsrow1: AtomsRow, atomsrow2: AtomsRow) -> bool:
    """Check the score AtomsRow similarity"""
    score = atoms_row_matcher_log(atomsrow1, atomsrow2)
    return np.mean(np.array(list(score.values())).astype(float))


class HashAtomsRow(AtomsRow):
    """Add equal method, the same site is the same atom."""

    @classmethod
    def from_atomsrow(cls, atomsrow: AtomsRow):
        """Return a copy."""
        atomss = cls(dct={"numbers": atomsrow.numbers})
        atomss.__dict__.update(atomsrow.__dict__)
        return atomss

    def to_atomsrow(self) -> AtomsRow:
        """Return a copy."""
        atomss = AtomsRow(dct={"numbers": self.numbers})
        atomss.__dict__.update(self.__dict__)
        return atomss

    def __hash__(self):
        return hash(self.__str__())

    def __eq__(self, ar2: AtomsRow, tol=2, resort=True):
        poa = np.round(self.positions, decimals=tol)
        pob = np.round(ar2.positions, decimals=tol)

        if resort:
            poa = np.sort(poa, axis=0)
            poa_index = np.argsort(poa, axis=0)
            pob = np.sort(pob, axis=0)
            pob_index = np.argsort(pob, axis=0)

            return (self.natoms == ar2.natoms and
                    (poa == pob).all() and (self.numbers[poa_index] == ar2.numbers[pob_index]).all())
        else:
            return (self.natoms == ar2.natoms and
                    (poa == pob).all() and (self.numbers == ar2.numbers).all())


def check_name_tup(name_pair):
    assert isinstance(name_pair, Iterable)
    for i in name_pair:
        assert len(i) == 2
        for j in i:
            assert isinstance(j, str) and j not in ["data", "_data", "key_value_pairs", "_keys"]


def atoms_row_rename(atomsrow: AtomsRow, name_pair=(("old_name1", "new_name1"),), check=True):
    """Rename the name in data or key_value_pairs of row."""
    if check:
        check_name_tup(name_pair)

    kvp = atomsrow.key_value_pairs
    data = atomsrow.get('data', {})

    olds = []

    for o, n in name_pair:
        if o in kvp:
            kvp[n] = kvp[o]
            del kvp[o]

        elif o in data:
            data[n] = data[o]
            del data[o]
        olds.append(o)

    [delattr(atomsrow, i) for i in olds if hasattr(atomsrow, i)]
    atomsrow._keys = list(kvp.keys())
    atomsrow.__dict__.update(kvp)
    atomsrow._data = {}
    atomsrow._data.update(data)
    return atomsrow
